"""Common stack layer

This module provides the `Layer` base class that provides a convenient way
to model a specific layer or group of layers that are part of a protocol stack.
It is the basic block of a protocol stack model used in WHAD.

A protocol layer does generally takes some data coming from its lower layer,
process them and forward it to an upper layer, as well as receiving data
from an upper layer, processing it and forwarding it to a lower layer. More
generally, it receives data from one or many layers, and send data to one or
more other layers. There is also a hierarchy in the stack model.

The `Layer` class
-----------------

The `Layer` class provides a convenient way to implement a protocol layer as
well as a protocol stack in its whole. Each layer defined using this class
can have sub-layers attached to it and act as a group of layers, send and
receive data from and to other layers. A derivative class called `ContextualLayer`
can be used to dynamically create some specific contexts that may be useful
when multiplexing/demultiplexing data.

Let's create a simple layer:

``` python
from whad.common.stack import Layer, alias

@alias('eth')
class EthLayer(Layer):
    pass
```

This `EthLayer` class uses the `alias` decorator to specify a text that will be
associated with this class. Thias alias will be used to reference this layer later.
By default, when a protocol stack is instantiated, an instance of each declared
class is automatically created and referenced by the corresponding alias.

Let's add an IP layer:

``` python
from whad.common.stack import Layer, alias

@alias('eth')
class EthLayer(Layer):
    pass

@alias('ip')
class IpLayer(Layer):
    pass
```

Now we have two layers defined, but they are totally alone and not connected.
Let's add a last layer that will handle the physical link:

``` python
@alias('phy')
class PhyLayer(Layer):
    pass
```

This physical layer will be our main layer for our stack, and will send each
raw packet received to the `EthLayer` class and receive as well packets to
send back on the physical link from the latter.

Layer messaging
---------------

Data flow between layers is managed by the `Layer` class that provides a very
simple communication mechanism. In fact, any declared layer that belongs to
a stack can send data to any other layer by simply using the `send()` method.
To receive data, a specific function decorator `source` allows the user to
specify which method has to be used to process messages coming from a specific
layer.

Let's add our dataflow into our model:

``` python
from whad.common.stack import Layer, alias, source

@alias('eth')
class EthLayer(Layer):

    @source('phy')
    def on_phy_packet_received(self, packet):
        '''Process received packet'''
        # some processing here
        ip_packet = unpack(packet)

        # forward to IP layer at some point
        self.send('ip', ip_packet)

    @source('ip')
    def on_transmit_ip_packet(self, ip_packet):
        '''Process IP packet to send'''
        eth_packet = pack(ip_packet)

        # send back this packet to phy
        self.send('phy', eth_packet)

@alias('ip')
class IpLayer(Layer):

    @source('eth')
    def on_ip_packet_received(self, packet):
        '''Process incoming IP packet'''
        # Process packet ...
        response = process_packet(packet)
        if response is not None:
            # Send back packet if required
            self.send('eth', response)

@alias('phy')
class PhyLayer(Layer):

    @source('eth')
    def on_transmit_eth_packet(self, packet):
        '''Transmit packet on the network'''
        # Some code here to effectively send the packet on the physical link
        pass

    def on_receive_phy_packet(self, packet):
        '''Received packet from the network'''
        # Send packet to our Ethernet layer
        self.send('eth', packet)
```

Using the above code, when the physical layer class `PhyLayer` is sending a
received raw packet to the ethernet layer (`EthLayer`), the data is automatically
routed and the method `on_phy_packet_received()` of the `EthLayer` is called
with the specified data.

Of course, it is possible for a layer to send data to any layer, the dataflow is
totally flexible.


Contextual layers
-----------------

Often, a layer is in charge of decapsulating and encapsulating data based on a
specific context. In our stack model, this type of layer is called a contextual
layer.

In our stack model, a contextual layer must inherit from `ContextualLayer` and
be instantiated to hold its own context. The lower layer must then instantiate
a dedicated contextual layer corresponding to a specific context, and dispatch
the incoming packets/messages to the correct contextual layer, thus performing
the mux/demux operation.
"""

def convert_layer_structure(structure):
    """Convert a layer structure into a GV DOT cluster
    """
    output = ''

    if structure['instanciable']:
        output += 'subgraph cluster_%s {\n' % structure['name']
        output += 'style=filled;\ncolor=lightgrey;\n'

    # Declare root node
    output += 'node [label="%s", fontsize=12] %s;\n' % (
        structure['name'],
        structure['name']
    )

    if len(structure['sublayers']) > 0:
        for sublayer in structure['sublayers']:
            output += convert_layer_structure(sublayer)

    if structure['instanciable']:
        #output += 'label="%s";\n' % structure['name']
        output += '}\n'

    return output

def generate_links(structure):
    """
    """

    links = []

    # Add emitters for current node
    for emitter in structure['emitters']:
        links.append((emitter, structure['name']))

    # Add emitters for all sublayers
    for sublayer in structure['sublayers']:
        links.extend(generate_links(sublayer))

    return links



class source(object):
    """Layer method decorator to perform source matching.

    Source matching can be done on layer name and optionally a tag (some
    user-defined text used to represent a specific state or operation).

    This tag is used to dispatch the incoming data to the correct callback,
    each callback accepting additional named arguments as specified by
    the implementer.
    """

    def __init__(self, source, tag='default', contextual=False):
        self.__source = source
        self.__tag = tag
        self.__contextual = contextual

    def __call__(self, func):
        """Manage source matching.
        """
        if hasattr(func, 'match_sources'):
            sources = getattr(func, 'match_sources')
            if isinstance(sources, dict):
                if self.__source not in sources:
                    # Add source if not already present
                    sources[self.__source] = [self.__tag]
                elif self.__tag not in sources[self.__source]:
                    sources[self.__source].append(self.__tag)
        else:
            func.match_sources = {self.__source: [self.__tag]}
            func.is_contextual = self.__contextual
        return func

class instance(source):
    """Instance has basically the same behavior as the source decorator,
    but forces it to include the instance reference.
    """
    def __init__(self, source, tag='default'):
        super().__init__(source, tag=tag, contextual=True)


class alias(object):
    """Layer class decorator to specify layer text alias.
    """

    def __init__(self, name):
        self.__name = name

    def __call__(self, clazz):
        clazz.alias = self.__name
        return clazz


class state(object):
    def __init__(self, state_class):
        self.__state_class = state_class

    def __call__(self, clazz):
        clazz.state_class = self.__state_class
        return clazz

class LayerState(object):
    """Stack layer state database

    Define fields names in FIELDS.
    """

    def __init__(self):
        """Populate database.
        """
        self.__db = {}
        for prop in dir(self):
            prop_obj = getattr(self, prop)
            if not prop.startswith('_') and not callable(prop_obj):
                self.__db[prop] = prop_obj

    def __getattr__(self, property):
        if property in self.__db:
            return self.__db[property]
        else:
            raise AttributeError

    def __setattr__(self, property, value):
        if property.startswith('_'):
            super(LayerState, self).__setattr__(property, value)
        else:
            self.__db[property] = value

    def to_dict(self):
        return self.__db

    def from_dict(self, values):
        for prop in values:
            self.__db[prop] = values[prop]


@state(LayerState)
class Layer(object):
    """
    Basic stack layer.
    """

    @classmethod
    def instantiable(cls):
        return False

    @classmethod
    def find(cls, alias):
        """Find a sub-layer class based on its alias
        """
        # First look into our sub-layers
        if hasattr(cls, 'LAYERS'):
            if alias in cls.LAYERS:
                return cls.LAYERS[alias]

            # If not found, propagate to our sub-layer classes
            for layer in cls.LAYERS:
                result = cls.LAYERS[layer].find(alias)
                if result is not None:
                    return result
        return None

    @classmethod
    def add(cls, clazz, input=False):
        """Add a sub-layer class.
        """
        # First we inject a LAYERS attribute into the class
        layers_prop_name = 'LAYERS'
        if not hasattr(cls, layers_prop_name):
            setattr(cls, layers_prop_name, {})
        class_layers = getattr(cls, layers_prop_name)

        # Then an input boolean field
        if input:
            setattr(cls, 'ENTRY_LAYER', clazz.alias)

        # Register a layer based on its alias
        if hasattr(clazz, 'alias'):
            if clazz.alias in cls.LAYERS:
                cls.LAYERS[clazz.alias] = clazz
            else:
                cls.LAYERS[clazz.alias] = clazz

    @classmethod
    def remove(cls, clazz):
        """Remove a sub-layer class.
        """
        layers_prop_name = 'LAYERS'
        if hasattr(cls, layers_prop_name):
            class_layers = getattr(cls, layers_prop_name)
            if clazz.alias in class_layers:
                del class_layers[clazz.alias]

    def __init__(self, parent=None, layer_name=None, options={}):
        self.__parent = parent
        self.__layer_name = layer_name
        self.__state = self.state_class()
        self.__layers = {}
        self.__layer_cache = {}
        self.__options = options
        self.__monitor_callbacks = []

        # Cache our message handlers
        self.__handlers = {}
        methods = []
        for prop in dir(self):
            try:
                prop_obj = getattr(self, prop)
                if callable(prop_obj):
                    methods.append(prop_obj)
            except AttributeError as att_err:
                pass

        for method in methods:
            if hasattr(method, 'match_sources') and isinstance(getattr(method, 'match_sources'), dict):
                match_sources = getattr(method, 'match_sources')
                for source in match_sources:
                    tags = match_sources[source]
                    for tag in tags:
                        handler_key = '%s:%s'%(source,tag)
                    self.__handlers[handler_key] = method

        # Call configure to set up options
        self.configure(options)

        # Populate all the sub-layers, if any.
        if hasattr(self, 'LAYERS'):
            self.populate(options)

    def populate(self, options={}):
        """Sub-layers instanciation.

        We instantiate each layer and register these instances into our object.
        """
        self.__options = options

        # Define layers and default context.
        for layer in self.LAYERS.keys():
            if not self.LAYERS[layer].instantiable():
                layer_inst = self.create_layer(self.LAYERS[layer], layer)
                if layer in options:
                    layer_inst.configure(options[layer])

    def instantiate(self, contextual_clazz):
        """Instantiate a contextual layer.
        """
        # Make sure the class inherits from `ContextualLayer` class
        if issubclass(contextual_clazz, ContextualLayer):
            # Build instance number
            if hasattr(contextual_clazz, 'INSTCOUNT'):
                instcount = getattr(contextual_clazz, 'INSTCOUNT')
                instcount += 1
            else:
                setattr(contextual_clazz, 'INSTCOUNT', 0)
                instcount = 0
            instance_name = '%s#%d' % (contextual_clazz.alias, instcount)

            # Create layer with this new instance name.
            return self.create_layer(contextual_clazz, instance_name)
        else:
            return None

    def create_layer(self, layer_class, inst_name):
        """Create a layer and registers it into our list of layers.
        """
        layer_options = self.options[layer_class.alias] if layer_class.alias in self.options else {}
        self.__layers[inst_name] = layer_class(self, inst_name, options=layer_options)
        return self.__layers[inst_name]

    def destroy(self, layer_instance):
        '''Remove an instantiated layer from our known layers.
        '''
        if layer_instance.name in self.__layers:
            del self.__layers[layer_instance.name]

    def register_monitor_callback(self, callback):
        '''Register a callback to monitor messages sent between layers.
        '''
        if callback not in self.__monitor_callbacks:
            self.__monitor_callbacks.append(callback)

    def unregister_monitor_callback(self, callback):
        '''Unregister a previously registered callback.
        '''
        if callback in self.__monitor_callbacks:
            self.__monitor_callbacks.remove(callback)

    def monitor_message(self, source, destination, data, tag='default', **kwargs):
        for monitor in self.__monitor_callbacks:
            monitor(source, destination, data, tag=tag, **kwargs)

    def has_layer(self, name):
        """Check if layer has a specific sublayer.
        """
        # First, we check if it is one of our sublayers
        if name in self.layers:
            return True
        else:
            # Check if one of our sublayer has this layer
            for layer in self.layers:
                if self.layers[layer].has_layer(name):
                    return True
        return False

    def has_handler(self, source, tag='default'):
        """Check if this layer has a registered method to process messages coming from a specific source/tag.
        """
        return (self.get_handler(source, tag=tag) is not None)

    def get_handler(self, source, tag='default'):
        """Retrieve the registered handler for a given source and tag (if any).
        """
        handler_key = '%s:%s'%(source, tag)
        if handler_key in self.__handlers:
            return self.__handlers[handler_key]

        # If not found, fall back on 'default' tag
        handler_key = '%s:%s'%(source, 'default')
        if handler_key in self.__handlers:
            return self.__handlers[handler_key]

        # Not found
        return None

    def get_layer(self, name, children_only=False):
        """Retrieve a specific layer based on its name.
        """
        # Are we the target layer ?
        if name == self.alias:
            # Return ourself :)
            return self

        # Do we have this layer in cache ?
        if name in self.__layer_cache:
            return self.__layer_cache[name]
        else:
            # First, we check if it is one of our sublayers
            if name in self.layers:
                return self.layers[name]
            else:
                # Check if one of our sublayer has this layer (children only)
                for layer in self.layers:
                    if not self.layers[layer].instantiable():
                        result = self.layers[layer].get_layer(name, children_only=True)
                        if result is not None:
                            # Found the layer, save in cache and return it
                            self.__layer_cache[name] = result
                            return result

                # If not, we ask our parent to get it
                if not children_only and self.__parent is not None:
                    layer = self.__parent.get_layer(name)

                    # Save layer in cache
                    if layer is not None:
                        self.__layer_cache[name] = layer

                    # Return layer
                    return layer

            # If anyone has this layer, it does not exist
            return None

    def get_entry_layer(self):
        """Return the group entry layer.
        """
        if hasattr(self, 'ENTRY_LAYER'):
            return getattr(self, 'ENTRY_LAYER')
        else:
            return None

    @property
    def name(self):
        return self.__layer_name if self.__layer_name is not None else self.alias

    @property
    def parent(self):
        return self.__parent

    @property
    def state(self):
        return self.__state

    @property
    def layers(self):
        return self.__layers

    @property
    def options(self):
        return self.__options

    @classmethod
    def list_emitters(cls):
        """Find sublayers that send messages to the specified layer.
        """
        emitters = []

        # First, loop on our own methods to find the sources we are using.
        methods = [getattr(cls, prop) for prop in dir(cls) if callable(getattr(cls, prop))]
        for method in methods:
            if hasattr(method, 'match_sources') and isinstance(getattr(method, 'match_sources'), dict):
                match_sources = getattr(method, 'match_sources')
                for _source in match_sources:
                    if _source not in emitters:
                        emitters.append(_source)

        return emitters

    def get_message_handler(self, source):
        """Find the message handler associated with the source
        """
        methods = [getattr(self, prop) for prop in dir(self) if callable(getattr(self, prop))]
        for method in methods:
            if hasattr(method, 'match_sources') and isinstance(getattr(method, 'match_sources'), dict):
                match_sources = getattr(method, 'match_sources')
                if source in match_sources:
                    if tag in match_sources[source]:
                        return method

    def __get_layer_handler_by_source(self, layer, source, tag='default'):
        """Find the method associated with a source for a given layer.
        """
        layer = self.get_layer(layer)
        if layer is None:
            return None
        else:
            methods = [getattr(layer, prop) for prop in dir(layer) if callable(getattr(layer, prop))]
            for method in methods:
                if hasattr(method, 'match_sources') and isinstance(getattr(method, 'match_sources'), dict):
                    match_sources = getattr(method, 'match_sources')
                    if source in match_sources:
                        if tag in match_sources[source]:
                            return method

        # if tag is not default, try again with 'default'
        for method in methods:
            if hasattr(method, 'match_sources') and isinstance(getattr(method, 'match_sources'), dict):
                match_sources = getattr(method, 'match_sources')
                if source in match_sources:
                    if 'default' in match_sources[source]:
                        return method
        return None

    def send(self, destination, data, tag='default', **kwargs):
        """Send a message to the corresponding layer.
        """
        return self.send_from(self.name, destination, data, tag=tag, **kwargs)

    def send_from(self, source, destination, data, tag='default', **kwargs):
        """Dispatch data from source to destination, with an optional tag
        and arguments.
        """
        # If source name has a '#' in it, then it is an instance of a
        # contextual layer and we must remove this to route the message.
        if '#' in source:
            idx = source.find('#')
            source_layer = source[:idx]
        else:
            source_layer = source

        # notify monitors
        self.monitor_message(source, destination, data, tag=tag, **kwargs)

        # Find the target layer object
        target_layer = self.get_layer(destination)
        if target_layer is not None:
            # Then we search the corresponding handler for our source
            handler = target_layer.get_handler(source_layer, tag)
            if handler is not None:
                if handler.is_contextual:
                    handler(source, data, **kwargs)
                else:
                    handler(data, **kwargs)
            else:
                print('[oops] No handler found in layer %s to process messages from %s' % (destination, source))
        else:
            print('[oops] layer %s does not exist' % destination)


    def __getitem__(self, name):
        """Array-like behavior to get a specific layer.
        """
        return self.get_layer(name)

    def configure(self, options):
        """Configure callback.

        Override this method to configure the layer when the stack is instantiated.
        """
        pass

    def save(self):
        """Return this layer saved state.
        """
        sublayers = {}
        for layer in self.layers:
            sublayers[layer] = self.layers[layer].save()

        layer_state = {
            'name': self.name,
            'state': self.__state.to_dict(),
            'sublayers': sublayers
        }

        return layer_state

    def load(self, state):
        """Set this layer properties dictionnary (used to load state).
        """
        # First, populate our state
        assert(state['name'] == self.name)
        self.state.from_dict(state['state'])

        # Populate our sublayers, instantiate contextual layers and set their state
        for sublayer in state['sublayers']:
            # If sublayer is not a contextual layer, set its state
            if '#' not in sublayer:
                sublayer_obj = self.get_layer(sublayer)
                sublayer_obj.load(state['sublayers'][sublayer])
            else:
                # sublayer class
                idx = sublayer.index('#')
                sublayer_class = sublayer[:idx]

                # instantiate and initialize
                sublayer_obj = self.create_layer(self.LAYERS[sublayer_class], sublayer)
                sublayer_obj.load(state['sublayers'][sublayer])

    @classmethod
    def get_structure(cls):
        """Retrieve layer structure.
        """


        # Populate sublayer structure
        sublayers_structure = []
        if hasattr(cls, 'LAYERS'):
            # Loop on all sublayers
            for sublayer in cls.LAYERS:
                sublayers_structure.append(cls.LAYERS[sublayer].get_structure())

        # Return our structure
        structure = {
            'name': cls.alias,
            'instanciable': cls.instantiable(),
            'emitters': cls.list_emitters(),
            'sublayers':sublayers_structure,
        }

        return structure

    @classmethod
    def export(cls, output_file=None):
        """Export to graphviz file.
        """
        structure = cls.get_structure()

        output = 'digraph T {\n'
        output += 'rankdir=LR;\n'

        # Walk the structure to extract the nodes/subgraphs
        output += convert_layer_structure(structure)

        # Walk the structure and extract the links
        links = generate_links(structure)

        for source, destination in links:
            output += '%s -> %s;\n' % (source, destination)

        output += '}'

        if output_file is not None:
            # Write to file
            with open(output_file, 'w') as f:
                f.write(output)

        return output


class ContextualLayer(Layer):
    """This layer is not automatically loaded when the stack model is created
    and must be instantiated specifically by another layer.
    """

    @classmethod
    def instantiable(cls):
        return True



from whad.zigbee.stack.apl.zdo.discovery import ZDODeviceAndServiceDiscovery
from whad.zigbee.stack.apl.zdo.discovery.clusters import ZDODeviceAnnce, ZDONodeDescReq, \
    ZDONodeDescRsp, ZDONWKAddrReq, ZDONWKAddrRsp, ZDOIEEEAddrReq, ZDOIEEEAddrRsp, \
    ZDOActiveEPReq, ZDOActiveEPRsp, ZDOSimpleDescReq, ZDOSimpleDescRsp

from whad.zigbee.stack.apl.zdo.network import ZDONetworkManager
from whad.zigbee.stack.apl.zdo.security import ZDOSecurityManager
from whad.zigbee.stack.apl.zdo.database import ConfigurationDatabase
from whad.zigbee.stack.apl.application import ApplicationObject

import logging

logger = logging.getLogger(__name__)

class ZigbeeDeviceObjects(ApplicationObject):
    """
    Application object representing Zigbee Device Objects.

    This app is the main entry point for managing the stack. It handles:
     - a specific database, storing the current configuration
     - the ZDO Network Manager, handling the network operations
     - the ZDO Security Manager, handling the security related operations
     - the ZDO Devices and Discovery Manager, handling the identification of
       surrounding devices and the discovery process of their apps & services.
    """
    def __init__(self):
        # Initialize the configuration database
        self.configuration = ConfigurationDatabase()

        # Initializes the managers
        self.security_manager = ZDOSecurityManager(self)
        self.network_manager = ZDONetworkManager(self)
        self.device_and_service_discovery = ZDODeviceAndServiceDiscovery(self)

        # Initialize ZDP clusters
        self.setup_clusters()

        # Configure application parameters
        super().__init__(
            "zdo",  # name
            0x0000, # profile ID
            0x0000, # device ID
            device_version=0,
            input_clusters=[
                self.clusters["node_desc_req"],
                self.clusters["node_desc_rsp"],
                self.clusters["ieee_addr_req"],
                self.clusters["nwk_addr_rsp"],
                self.clusters["ieee_addr_rsp"],
                self.clusters["active_ep_req"],
                self.clusters["active_ep_rsp"],
                self.clusters["simple_desc_req"],
                self.clusters["simple_desc_rsp"],
            ],
            output_clusters=[
                self.clusters["nwk_addr_req"],
                self.clusters["active_ep_req"],
                self.clusters["active_ep_rsp"],
                self.clusters["ieee_addr_req"],
                self.clusters["device_annce"],
                self.clusters["node_desc_req"],
                self.clusters["node_desc_rsp"],
                self.clusters["simple_desc_req"],
                self.clusters["simple_desc_rsp"],
            ]
        )


    def setup_clusters(self):
        self.clusters = {
            "device_annce"      : ZDODeviceAnnce(self.device_and_service_discovery),
            "node_desc_req"     : ZDONodeDescReq(self.device_and_service_discovery),
            "node_desc_rsp"     : ZDONodeDescRsp(self.device_and_service_discovery),
            "nwk_addr_req"      : ZDONWKAddrReq(self.device_and_service_discovery),
            "nwk_addr_rsp"      : ZDONWKAddrRsp(self.device_and_service_discovery),
            "ieee_addr_req"     : ZDOIEEEAddrReq(self.device_and_service_discovery),
            "ieee_addr_rsp"     : ZDOIEEEAddrRsp(self.device_and_service_discovery),
            "active_ep_req"     : ZDOActiveEPReq(self.device_and_service_discovery),
            "active_ep_rsp"     : ZDOActiveEPRsp(self.device_and_service_discovery),
            "simple_desc_req"   : ZDOSimpleDescReq(self.device_and_service_discovery),
            "simple_desc_rsp"   : ZDOSimpleDescRsp(self.device_and_service_discovery),
        }

    def initialize(self):
        """
        Initializes the ZDO managers.
        """
        self.network_manager.initialize()

    def start(self):
        """
        Starts the ZDO managers.
        """
        self.network_manager.startup()

    def configure(self, attribute_name, attribute_value):
        """
        Configures the value of a specific attribute in the database.
        """
        return self.configuration.set(attribute_name, attribute_value)

    def on_join(
                            self,
                            network_address,
                            extended_address,
                            capability_information,
                            rejoin=False,
                            secure_rejoin=False
    ):
        """
        Forwards new device join notification.
        """
        self.security_manager.send_transport_key(
            network_address
        )

    def on_transport_key(
                            self,
                            source_address,
                            standard_key_type,
                            transport_key_data
    ):
        """
        Forwards the transport key to the security manager.
        """
        self.security_manager.on_transport_key(
            source_address,
            standard_key_type,
            transport_key_data
        )

    # Helpers
    @property
    def nwk_management(self):
        """
        Shortcut for accessing the NWK management service.
        """
        return self.manager.get_layer('nwk').get_service('management')

    @property
    def aps_management(self):
        """
        Shortcut for accessing the APS management service.
        """
        return self.manager.get_layer('aps').get_service('management')



"""WHAD Protocol PHY modulation messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.protocol.phy.phy_pb2 import GetSupportedFrequenciesCmd
from ..message import pb_bind, PbFieldBool, PbMessageWrapper
from . import PhyDomain

@pb_bind(PhyDomain, 'get_supported_freq', 1)
class GetSupportedFreqs(PbMessageWrapper):
    """PHY Get supported frequencies message
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.phy.get_supported_freq.CopyFrom(GetSupportedFrequenciesCmd())

@pb_bind(PhyDomain, 'set_freq', 1)
class SetFreq(PbMessageWrapper):
    """PHY SetFreq modulation message
    """

    frequency = PbFieldBool('phy.set_freq.frequency')

@pb_bind(PhyDomain, 'supported_freq', 1)
class SupportedFreqRanges(PbMessageWrapper):
    """PHY SupportedFrequencyRanges message
    """

    ranges = PbFieldBool('phy.supported_freq.frequency_ranges')

    def add(self, start: int, end: int):
        """Add a supported frequency range to this message

        :param start: Start frequency of this range
        :type start: int
        :param end: End frequency of this range
        :type end: int
        """
        range = self.ranges.add()
        range.start = start
        range.end = end


"""Logitech Unifying mouse tool

This utility provides multiple features targeting Logitech Unifying mice:
- mouse movement and click logging
- interactive mouse and click injection (taking control over a wireless mouse)
- mouse and click injection from stdin
"""
import sys
import time
import logging

# Whad dependencies
from pynput import mouse
from whad.cli.app import CommandLineSink, run_app
from whad.unifying.connector import Mouse, Mouselogger, ESBAddress
from whad.unifying.stack.constants import ClickType
from whad.esb.exceptions import InvalidESBAddressException

# Logging
logger = logging.getLogger(__name__)

TOOL_DESCRIPTION="""This tool synchronizes with an existing mouse and then logs all mouse moves and
button presses or inject moves and button presses.
"""

class UniMouseApp(CommandLineSink):
    """Logitech Unifying mouse CLI app
    """

    mouse_coords = None

    def __init__(self):
        """Application uses an interface and has no commands.
        """
        super().__init__(
            description=TOOL_DESCRIPTION,
            interface=True,
            commands=False
        )

        # Target device address option
        self.add_argument(
            '-a',
            '--address',
            metavar='ADDRESS',
            dest='address',
            default=None,
            help="Target mouse with address ADDRESS"
        )

        # Mouse logging raw output
        self.add_argument(
            '-r',
            '--raw',
            dest='raw',
            action="store_true",
            default=False,
            help="Log mouse output in raw format"
        )

        # Mouse duplication option
        self.add_argument(
            '-d',
            '--duplicate',
            dest='duplicate',
            action='store_true',
            default=False,
            help="If set, send your current mouse moves and clicks to the target mouse dongle"
        )

    def run(self):
        """Override App's run() method to handle scripting feature.
        """
        try:
            # Launch pre-run tasks
            self.pre_run()

            # We need to have an interface specified
            if self.interface is not None:
                # If an address is not provided, we cannot go further
                if self.args.address is None:
                    # Start scanning
                    self.error("You must specify a target mouse address with -a/--address")
                else:
                    # If an address is provided, we check its format and start
                    # sniffing packets from this device
                    try:
                        # Check address format
                        addr = ESBAddress(self.args.address)

                        # If -r option is set, sync and replicate our mouse moves
                        if self.args.duplicate:
                            self.duplicate_mouse(addr)
                        else:
                            # If stdin is piped, then we expect some mouse moves and clicks
                            # coming from stdin
                            if self.is_stdin_piped():
                                self.duplicate_stdin(addr)
                            else:
                                # No action specified, simply log mouse events
                                self.log_mouse(addr)
                    except InvalidESBAddressException:
                        # Invalid device address
                        self.error('Target address does not match the expected format !')

            else:
                # Missing interface.
                self.error('You need to specify an interface with option --interface.')

        except KeyboardInterrupt:
            self.warning('wuni-mouse stopped (CTL-C)')

        # Launch post-run tasks
        self.post_run()

    def duplicate_mouse(self, address: ESBAddress):
        """Duplicate current mouse to the target wireless mouse dongle.
        """
        # Sync with target mouse
        connector = Mouse(self.interface)
        connector.start()
        connector.channel = 5
        connector.address =  str(address)
        connector.synchronize()

        print('Mouse found and locked, duplicating mouse moves and clicks (hit CTL-C to stop)')

        def on_move(x, y):
            """Forward captured mouse movement to target mouse.
            """
            mouse_coords = UniMouseApp.mouse_coords
            if mouse_coords is None:
                UniMouseApp.mouse_coords = (x,y)
            else:
                orig_x, orig_y = mouse_coords
                dx, dy = x - orig_x, y - orig_y
                displacement = dx*dx+dy*dy
                if displacement >= 49:
                    UniMouseApp.mouse_coords = (x,y)
                    connector.move(dx, dy)

        def on_scroll(x, y, dx, dy):
            """Forward captured wheel movement to target mouse.
            """
            if dx>0:
                connector.wheel_down()
            elif dx<0:
                connector.wheel_up()
            if dy>0:
                connector.wheel_right()
            elif dy<0:
                connector.wheel_left()

        def on_click(x, y, button, pressed):
            if button == mouse.Button.left:
                if not pressed:
                    connector.left_click()
            elif button == mouse.Button.right:
                if not pressed:
                    connector.right_click()
            elif button == mouse.Button.middle:
                if not pressed:
                    connector.middle_click()

        # Catch mouse events and dispatch to callbacks
        try:
            with mouse.Listener(
                on_move=on_move,
                on_click=on_click,
                on_scroll=on_scroll) as listener:
                listener.join()
        except KeyboardInterrupt:
            self.warning('Interrupted by user (CTL-C)')

    def duplicate_stdin(self, address: ESBAddress):
        """Duplicate stdin moves and clicks to the target wireless mouse dongle.

        Read lines from stdin and inject corresponding moves. Timing need to be
        handled bythe program sending mouse moves (and clicks) on stdin.

        Linme format is the following:

          X,Y,WHEEL_X,WHEEL_Y,BTNS

        with X and Y decimal integers representing the mouse movement, WHEEL_X
        and WHEEL_Y decimal integers representing the mouse wheel movement and
        BTNS being a series of letters ffrom "L", "R" and "M" corresponding to
        a click event to send.
        """
        # Sync with target mouse
        connector = Mouse(self.interface)
        connector.start()
        connector.channel = 5
        connector.address =  str(address)
        connector.synchronize()

        print('Mouse found and locked, sending moves received on stdin...')

        # Read lines from stdin
        for line in sys.stdin:
            try:
                move = line.split(',')
                if len(move) == 5:
                    dx,dy,wx,wy,btns = move

                    # Extract move delta
                    dx = int(dx) if len(dx) > 0 else 0
                    dy = int(dy) if len(dy) > 0 else 0
                    wx = int(wx) if len(wx) > 0 else 0
                    wy = int(wy) if len(wy) > 0 else 0

                    # Cap values ([-2047, 2047] for dx/dy and
                    # [-127,127] for wx/wy
                    if dx > 2047:
                        dx = 2047
                    elif dx < -2047:
                        dx = -2047
                    if dy > 2047:
                        dy = 2047
                    elif dy < -2047:
                        dy = -2047

                    # Send mouse move if any
                    # (0,0) means no move !
                    if dx != 0 or dy != 0:
                        connector.move(dx, dy)

                    # Send wheel moves (if any)
                    if wx > 0:
                        connector.wheel_down()
                    elif wx < 0:
                        connector.wheel_up()
                    if wy > 0:
                        connector.wheel_right()
                    elif wy < 0:
                        connector.wheel_left()

                    # Send clicks if any
                    btns = btns.upper()
                    if 'R' in btns:
                        connector.right_click()
                    if 'L' in btns:
                        connector.left_click()
                    if 'M' in btns:
                        connector.middle_click()

            except ValueError:
                self.warning('Invalid value in mouse move')

        # Wait for connector to send message
        time.sleep(1)
        connector.stop()


    def log_mouse(self, address: ESBAddress):
        """Log mouse moves and button presses.
        """
        connector = Mouselogger(self.interface)
        connector.address = str(address)
        connector.scanning = True
        connector.decrypt = True

        buttons_state = [0, 0, 0]

        connector.start()
        for delta, wheel, buttons in connector.stream():
            buttons_event = []
            # detect any button state change
            if buttons & ClickType.LEFT > 0:
                if buttons_state[0] == 0:
                    buttons_event.append("left button pressed")
                buttons_state[0] = 1
            else:
                if buttons_state[0] == 1:
                    buttons_event.append("left button released")
                buttons_state[0] = 0
            if buttons & ClickType.RIGHT > 0:
                if buttons_state[1] == 0:
                    buttons_event.append("right button pressed")
                buttons_state[1] = 1
            else:
                if buttons_state[1] == 1:
                    buttons_event.append("right button released")
                buttons_state[1] = 0
            if buttons & ClickType.MIDDLE > 0:
                if buttons_state[2] == 0:
                    buttons_event.append("middle button pressed")
                buttons_state[2] = 1
            else:
                if buttons_state[2] == 1:
                    buttons_event.append("middle button released")
                buttons_state[2] = 0

            wheel_events = []
            if wheel[0] != 0:
                wheel_events.append("wheel_x:"+str(wheel[0]))
            if wheel[1] != 0:
                wheel_events.append("wheel_y:"+str(wheel[1]))

            if len(buttons_event) > 0 or len(wheel_events) > 0:
                events = " | " + ", ".join(wheel_events + buttons_event)
            else:
                events = ""
            if not self.args.raw:
                print(f"Mouse move (dx:{delta[0]:d}, dy:{delta[1]:d}){events}")
            else:
                print(str(delta[0])+","+str(delta[1])+","+str(wheel[0])+","+str(wheel[1])+","+str(buttons))

def wuni_mouse_main():
    """Logitech Unifying mouse tool main routine.
    """
    app = UniMouseApp()
    run_app(app)



"""
This module provides some constants used by WHAD to communicate with the RFStorm firmware.
"""
from enum import IntEnum

# RFStorm Internal states
class RFStormInternalStates(IntEnum):
    """Internal state constants.
    """
    NONE                            = 0
    SNIFFING                        = 1
    PROMISCUOUS_SNIFFING            = 2
    PROMISCUOUS_GENERIC_SNIFFING    = 3
    TRANSMITTING                    = 4
    GENERIC_TRANSMITTING            = 5

# USB identifiers
class RFStormId(IntEnum):
    """RFStorm USB ID contstants
    """
    RFSTORM_ID_VENDOR    = 0x1915
    RFSTORM_ID_PRODUCT   = 0x0102

# USB Endpoints
class RFStormEndPoints(IntEnum):
    """RFStorm USB endpoints numbers
    """
    RFSTORM_COMMAND_ENDPOINT     = 0x01
    RFSTORM_RESPONSE_ENDPOINT    = 0x81

# USB Commands
class RFStormCommands(IntEnum):
    """RFStorm commands
    """
    RFSTORM_CMD_TRANSMIT = 0x04
    RFSTORM_CMD_SNIFF = 0x05
    RFSTORM_CMD_PROMISCUOUS = 0x06
    RFSTORM_CMD_TONE    = 0x07
    RFSTORM_CMD_TRANSMIT_ACK = 0x08
    RFSTORM_CMD_SET_CHANNEL = 0x09
    RFSTORM_CMD_GET_CHANNEL = 0x0a
    RFSTORM_CMD_ENABLE_LNA = 0x0b
    RFSTORM_CMD_TRANSMIT_GENERIC = 0x0C
    RFSTORM_CMD_PROMISCUOUS_GENERIC = 0x0D
    RFSTORM_CMD_RECV = 0x12

# RF Datarates
class RFStormDataRate(IntEnum):
    """RFStorm datarate constants
    """
    RF_250KBPS = 0
    RF_1MBPS = 1
    RF_2MBPS = 2

# Supported domains
class RFStormDomains(IntEnum):
    """RFStorm supported domains
    """
    RFSTORM_RAW_ESB = 0
    RFSTORM_UNIFYING = 1
    RFSTORM_PHY = 2



"""Att Exceptions
"""
class AttErrorCode:
    """ATT error codes
    """
    INVALID_HANDLE_VALUE = 0x01
    READ_NOT_PERMITTED = 0x02
    WRITE_NOT_PERMITTED = 0x03
    INVALID_PDU = 0x04
    INSUFFICIENT_AUTHENTICATION = 0x05
    UNSUPPORTED_REQUEST = 0x06
    INVALID_OFFSET = 0x07
    INSUFFICIENT_AUTHORIZATION = 0x08
    PREPARE_QUEUE_FULL = 0x09
    ATTR_NOT_FOUND = 0x0A
    ATTR_NOT_LONG = 0x0B
    INSUFFICIENT_ENC_KEY_SIZE = 0x0C
    INVALID_ATTR_VALUE_LENGTH = 0x0D
    UNLIKELY_ERROR = 0x0E
    INSUFFICIENT_ENC = 0x0F
    UNSUPPORTED_GROUP_TYPES = 0x10
    INSUFFICIENT_RES = 0x11

class AttError(Exception):
    """ATT generic exception
    """

    def __init__(self, request, handle):
        super().__init__()
        self.request = request
        self.handle = handle

class InvalidHandleValueError(AttError):
    """Exception raised when an invalid handle value is provided.
    """

class ReadNotPermittedError(AttError):
    """Exception raised when a read operation is performed while not permitted
    """

class WriteNotPermittedError(AttError):
    """Exception raised when a write operation is performed while not allowed
    """

class InvalidPduError(AttError):
    """Exception raised when an invalid PDU is processed
    """

class InsufficientAuthenticationError(AttError):
    """Authentication required
    """

class UnsupportedRequestError(AttError):
    """Exception raised when an unsupported request is received
    """

class InvalidOffsetError(AttError):
    """Exception raised when an invalid offset is provided to a read or write
    operation
    """

class InsufficientAuthorizationError(AttError):
    """Exception raised if an authorization is required
    """

class PrepareQueueFullError(AttError):
    """Raised when a prepare write queue is full.
    """

class AttributeNotFoundError(AttError):
    """Exception raised when an attribute cannot be found.
    """

class AttributeNotLongError(AttError):
    """Raised a read long operation is performed on a not-long attribute.
    """

class InsufficientEncryptionKeySize(AttError):
    """Raised when wrong key size is used
    """

class InvalidAttrValueLength(AttError):
    """Raised when an invalid length is used on an attribute
    """

class UnlikelyError(AttError):
    """Raised when an unlikely error is triggered
    """

class InsufficientEncryptionError(AttError):
    """Raised on insufficient encryption.
    """

class UnsupportedGroupTypesError(AttError):
    """Raised when an unsupported group type is requested.
    """

class InsufficientResourcesError(AttError):
    """Raised to notify an error of resources
    """

def error_response_to_exc(error_code, request, handle) -> AttError:
    """Convert error code to corresponding exception.

    :param error_code: ATT error code
    :type error_code: int
    :param request: Request that lead to this error code
    :type request: int
    :param handle: related handle
    :type handle: int
    :return: corresponding exception
    :rtype: AttError
    """
    resp_to_exc = {
        0x01: InvalidHandleValueError,
        0x02: ReadNotPermittedError,
        0x03: WriteNotPermittedError,
        0x04: InvalidPduError,
        0x05: InsufficientAuthenticationError,
        0x06: UnsupportedRequestError,
        0x07: InvalidOffsetError,
        0x08: InsufficientAuthorizationError,
        0x09: PrepareQueueFullError,
        0x0A: AttributeNotFoundError,
        0x0B: AttributeNotLongError,
        0x0C: InsufficientEncryptionKeySize,
        0x0D: InvalidAttrValueLength,
        0x0E: UnlikelyError,
        0x0F: InsufficientEncryptionError,
        0x10: UnsupportedGroupTypesError,
        0x11: InsufficientResourcesError
    }
    if error_code in resp_to_exc:
        return resp_to_exc[error_code](request, handle)
    else:
        return AttError(request, handle)



'''
ADPCM codec implementation.

*All* credits for this code goes to https://github.com/ShowerXu/python-adpcm/.
'''
from whad.scapy.layers.rf4ce import RF4CE_Vendor_MSO_Audio_Start_Request, \
		RF4CE_Vendor_MSO_Audio_Data_Notify, RF4CE_Vendor_MSO_Audio_Stop_Request, \
		RF4CE_Vendor_MSO_Audio
from struct import pack, unpack
import wave

resolution = 16

# Build geometric lookup table with base as 1.1, as used in IMA and OKI
STEP_TABLE = [7]
while STEP_TABLE[-1] <= (2**resolution) / 2:
		STEP_TABLE.append(1.1 * STEP_TABLE[-1])

# Trim last one which will be outside the valid range
STEP_TABLE = [int(round(x)) for x in STEP_TABLE[:-1]]

# Index table from IMA and OKI
INDEX_TABLE = [
		-1, -1, -1, -1, 2, 4, 6, 8
]

class ByteAdpcmEncoder:
	def __init__(self, start):
		self.predicted = start
		self.start_index = None
		self.index = None

	def _find_best_index(self, diff):
		best_step = 0
		best_step_diff = abs(STEP_TABLE[0] - diff)

		for i in range(1, len(STEP_TABLE)):
			cur_step_diff = abs(STEP_TABLE[i] - diff)
			if cur_step_diff < best_step_diff:
				best_step = i
				best_step_diff = cur_step_diff

		return best_step

	def encode(self, value):
		delta = value - self.predicted
		if delta < 0:
			isNeg = True
			delta = -delta
		else:
			isNeg = False

		if self.start_index is None:
			self.start_index = self.index = self._find_best_index(delta)

		index = self.index
		predicted = self.predicted
		step = STEP_TABLE[index]

		predicted_delta = 0
		encoded = 0

		if delta >= step:
			delta -= step
		predicted_delta += step
		encoded |= 0x4

		step = step >> 1
		if delta >= step:
			delta -= step
		predicted_delta += step
		encoded |= 0x2

		step >>= 1
		if delta >= step:
			predicted_delta += step
		encoded |= 0x1

		# Is this needed?
		step >>= 1
		predicted_delta += step

		index += INDEX_TABLE[encoded]
		if index < 0:
			index = 0
		elif index >= len(STEP_TABLE):
			index = len(STEP_TABLE) - 1
		self.index = index

		if isNeg:
			encoded |= 0x8
			predicted -= predicted_delta
		else:
			predicted += predicted_delta

		if predicted > 2**resolution:
			predicted = 2**resolution
		elif predicted < 0:
			predicted = 0
		self.predicted = predicted

		return encoded

class ByteAdpcmDecoder:
	def __init__(self, start_value, index):
		self.predicted = start_value
		self.index = index
		if index >= len(STEP_TABLE):
			raise ValueError('Invalid index')

	def decode(self, nibble):
		index = self.index
		step = STEP_TABLE[index]
		predicted = self.predicted

		predicted_delta = 0
		if nibble & 4:
			predicted_delta = step

		step >>= 1
		if nibble & 2:
			predicted_delta += step

		step >>= 1
		if nibble & 1:
			predicted_delta += step

		step >>= 1
		predicted_delta += step

		if nibble & 8:
			predicted -= predicted_delta
		else:
			predicted += predicted_delta

		if predicted > 2**resolution:
			predicted = 2**resolution
		elif predicted < 0:
			predicted = 0

		index += INDEX_TABLE[nibble & 0x7]
		if index < 0:
			index = 0
		elif index >= len(STEP_TABLE):
			index = len(STEP_TABLE) - 1

		self.index = index
		self.predicted = predicted
		return predicted



class ADPCM:
	def __init__(self, live_play=True, output_filename=None):
		self.output_filename = output_filename
		self.output_stream = None
		self.output_file = None
		self.live_play = live_play
		self.decoder = ByteAdpcmDecoder(0, 0)
		self.encoder = None

	def _open_stream(self, channels=1, rate=16000, sample_size=2, unsigned_sample_size=False):
		try:
			import pyaudio

			p = pyaudio.PyAudio()
			stream = p.open(
								format=pyaudio.get_format_from_width(
									sample_size,
									unsigned=unsigned_sample_size
								),
			                    channels=channels,
			                    rate=rate,
			                    output=True
			)
			return stream

		except ImportError:
			return None

	def _open_file(self, channels=1, rate=16000, sample_size=2, unsigned_sample_size=False):
		stream = wave.open(self.output_filename, 'wb')
		stream.setnchannels(channels)
		stream.setsampwidth(sample_size)
		stream.setframerate(rate)

		return stream

	def decode(self, samples):
		decoded_samples = b""
		for frame in samples:
			low = self.decoder.decode(frame >> 4)
			high = self.decoder.decode(frame & 0xF)
			try:
				decoded_samples += pack('H', low) + pack('H', high)
			except:
				pass
		return decoded_samples

	def convert_input_file(self, filename="/tmp/trololo.wav"):
		# buggy
		stream = wave.open(filename, "rb")
		nchannels = stream.getnchannels()
		sampwidth = stream.getsampwidth()
		rate = stream.getframerate()
		frames = stream.readframes(stream.getnframes())
		out = b""
		import audioop
		s = None
		(frag, ns) = audioop.lin2adpcm(frames,2,s)
		out += frag

		return ([
		RF4CE_Vendor_MSO_Audio_Start_Request(
			sample_rate = rate,
			#sample_rate=16000,
			resolution_bits=16,
			mic_channel_number=1,
			packet_size=84,
			interval=10,
			channel_number=3,
			duration=10
			)]
			+
			[
				RF4CE_Vendor_MSO_Audio_Data_Notify(
				samples=out[s:s+80]
				)
				for s in range(0, len(out), 80)
			]
		)

	def process_packet(self, packet):
		if RF4CE_Vendor_MSO_Audio_Start_Request in packet:
			if self.live_play:
				self.output_stream = self._open_stream()
			if self.output_filename:
				self.output_file = self._open_file()

		elif RF4CE_Vendor_MSO_Audio_Data_Notify in packet:
			if self.output_stream is not None or self.output_file is not None:
				decoded_samples = self.decode(packet.samples)
			if self.output_stream is not None:
				if self.live_play:
					self.output_stream.write(decoded_samples)
			if self.output_file is not None:
				self.output_file.writeframes(decoded_samples)

		elif RF4CE_Vendor_MSO_Audio_Stop_Request in packet or (RF4CE_Vendor_MSO_Audio in packet and packet.audio_cmd_id == 2):
			if self.output_stream is not None:
				if self.live_play:
					self.output_stream.close()
					self.output_stream = None
				if self.output_file is not None:
					self.output_file.close()



"""LoRaWAN helpers
"""
import re

from .exceptions import BadEuiFormat

class EUI(object):
    """Extended Unique Identifier
    """

    def __init__(self, eui : str):
        """Initialize an EUI
        """
        # check EUI format
        if re.match('^([0-9a-fA-F]{2}:){7}[0-9a-fA-F]{2}$', eui):
            values = eui.lower().split(':')[::-1]
            self.__packed = bytes([int(v, 16) for v in values])
        else:
            raise BadEuiFormat()

    @property
    def value(self) -> bytes:
        """Return EUI as a packed value (bytes)

        :return: bytes
        """
        return self.__packed

    def __repr__(self):
        return ':'.join(['%02x' % b for b in self.__packed][::-1])
    
    def __eq__(self, other) -> bool:
        """Compare two EUIs
        """
        return self.value == other.value




# This script is used to print out a UUID128 in different formats,
# which may appear in source code when searching sites like GitHub.
# Usage:
#   python3 UUID_Format_Permutation.py <UUID128 with or without dashes>
# Example:
#   python3 UUID_Format_Permutation.py 5052494D-2DAB-0341-6972-6F6861424C45
# Output:
#   Without dashes: 5052494D2DAB034169726F6861424C45
#   With dashes:    5052494D-2DAB-0341-6972-6F6861424C45
#   Little-endian C byte array: {0x45, 0x4C, 0x42, 0x61, 0x68, 0x6F, 0x72, 0x69, 0x41, 0x03, 0xAB, 0x2D, 0x4D, 0x49, 0x52, 0x50}
#   Little-endian C byte array (no spaces): 0x45,0x4C,0x42,0x61,0x68,0x6F,0x72,0x69,0x41,0x03,0xAB,0x2D,0x4D,0x49,0x52,0x50

import sys
import re

def normalize_uuid(uuid_str):
    # Remove dashes and convert to uppercase
    return re.sub(r'[^0-9A-Fa-f]', '', uuid_str).upper()

def with_dashes(uuid_str):
    # Insert dashes at standard UUID positions
    return f"{uuid_str[0:8]}-{uuid_str[8:12]}-{uuid_str[12:16]}-{uuid_str[16:20]}-{uuid_str[20:32]}"

def little_endian_bytes(uuid_str):
    # Convert hex string to bytes, reverse for little-endian
    b = bytes.fromhex(uuid_str)
    b_le = b[::-1]
    return b_le

def format_c_array(b_le, with_spaces=True):
    arr = [f"0x{byte:02X}" for byte in b_le]
    sep = ', ' if with_spaces else ','
    return sep.join(arr)

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Usage: python UUID_Format_Permutation.py <UUID128>")
        sys.exit(1)

    input_uuid = sys.argv[1]
    uuid_nodash = normalize_uuid(input_uuid)
    if len(uuid_nodash) != 32:
        print("Error: UUID must be 128 bits (32 hex digits).")
        sys.exit(1)

    uuid_dash = with_dashes(uuid_nodash)
    b_le = little_endian_bytes(uuid_nodash)

    print("Without dashes: " + uuid_nodash)
    print("Without dashes, endian-swapped: " + b_le.hex().upper()) # This is for if someone got their endianness wrong
    print("With dashes:    " + uuid_dash)
    print("Little-endian C byte array: {" + format_c_array(b_le, with_spaces=True) + "}")
    print("Little-endian C byte array (no spaces): " + format_c_array(b_le, with_spaces=False))


from whad.zigbee.stack.apl.zcl.exceptions import ZCLAttributePermissionDenied, \
    ZCLAttributeNotFound

class ZCLAttribute:
    """
    This class represents a Zigbee Cluster Library attribute.
    """
    def __init__(self, name, value=None, permissions=["read", "write"]):
        self.name = name
        self.value = value
        self.permissions = permissions

class ZCLAttributes:
    """
    This class represents a database of Zigbee Cluster Library attributes.
    """
    def __init__(self):
        self.attributes = {}

    def add_attribute(self, id, name, value, permissions=['read', 'write']):
        """
        Adds an attribute in the database.
        """
        self.attributes[id] = ZCLAttribute(
            name=name,
            value=value,
            permissions=permissions
        )

    def read_by_id(self, id):
        """
        Reads an attribute value according to its identifier.
        """
        if id in self.attributes:
            attribute = self.attributes[id]
            if "read" in attribute.permissions:
                return attribute.value
            raise ZCLAttributePermissionDenied()
        raise ZCLAttributeNotFound()

    def read_by_name(self, name):
        """
        Reads an attribute value according to its name.
        """
        for attribute in self.attributes:
            if attribute.name == name:
                if "read" in attribute.permissions:
                    return attribute.value
                else:
                    raise ZCLAttributePermissionDenied()
        raise ZCLAttributeNotFound()

    def write_by_id(self, id, value):
        """
        Writes an attribute value according to its identifier.
        """
        if id in self.attributes:
            attribute = self.attributes[id]
            if "write" in attribute.permissions:
                attribute.value = value
            raise ZCLAttributePermissionDenied()
        raise ZCLAttributeNotFound()

    def write_by_name(self, name, value):
        """
        Writes an attribute value according to its name.
        """
        for attribute in self.attributes:
            if attribute.name == name:
                if "write" in attribute.permissions:
                    attribute.value = value
                else:
                    raise ZCLAttributePermissionDenied()
        raise ZCLAttributeNotFound()



from whad.unifying.connector import Sniffer
from whad.unifying.stack.constants import ClickType
from whad.unifying.hid import LogitechUnifyingMouseMovementConverter
from whad.scapy.layers.unifying import Logitech_Mouse_Payload

class Mouselogger(Sniffer):
    """
    Logitech Unifying Mouse logger interface for compatible WHAD device.
    """

    def __init__(self, device):
        super().__init__(device)

    def stream(self):
        for packet in super().sniff():
            hid_data = None
            if Logitech_Mouse_Payload in packet:
                movement = packet.movement

                converter = LogitechUnifyingMouseMovementConverter()
                x, y = converter.get_coordinates_from_hid_data(movement)
                button = ClickType(packet.button_mask)

                yield ((x,y), (packet.wheel_x, packet.wheel_y), button)



"""Physical layer domain exceptions
"""

class UnknownPhysicalLayer(Exception):
    def __init__(self):
        super().__init__()

class UnknownPhysicalLayerFunction(Exception):
    def __init__(self, missing_function):
        self.missing_function = missing_function
        super().__init__()

class UnsupportedFrequency(Exception):
    def __init__(self, frequency):
        super().__init__(self, frequency)

class NoModulation(Exception):
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return "NoModulation: No modulation provided"

class NoDatarate(Exception):
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return "NoDatarate: No datarate provided"

class NoFrequency(Exception):
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return "NoFrequency: No frequency provided"

class NoSyncWord(Exception):
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return "NoSyncWord: No synchronization word provided"

class NoEndianess(Exception):
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return "NoEndianness: No endianness provided"

class NoPacketSize(Exception):
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return "NoPacketSize: No packet size provided"
    
class InvalidParameter(Exception):
    def __init__(self, parameter):
        super().__init__()
        self.__parameter = parameter

    def __repr__(self):
        return "InvalidParameter: provided %s parameter is not valid" % (self.__parameter)
    
class ScheduleFifoFull(Exception):
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return "Schedule packets FIFO is full"



"""WHAD Protocol ESB address messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
#from whad.protocol.esb.esb_pb2 import GetSupportedFrequenciesCmd
from ..message import pb_bind, PbFieldBytes, PbFieldBool, PbFieldArray, PbMessageWrapper
from . import EsbDomain

@pb_bind(EsbDomain, 'set_node_addr', 1)
class SetNodeAddress(PbMessageWrapper):
    """ESB SetNodeAddress message
    """

    address = PbFieldBytes('esb.set_node_addr.address')



from dataclasses import dataclass, field
from whad.common.sniffing import SniffingEvent

@dataclass
class SnifferConfiguration:
    """
    Configuration for sniffing a RF4CE communication.

    :param channel: select the channel to sniff (c)
    :param decrypt: indicate if decryption is enabled (d)
    :param keys: provide decryption keys (k)
    :param audio: decode audio stream (audio)
    :param pairing: sniff pairing procedure and break key if possible(p)
    :param addresses: provide node extended addresses (a)

    """
    channel : int = 15
    decrypt : bool = False
    pairing : bool = False
    audio : bool = False
    audio_file : str = ""
    keys : list = field(default_factory=lambda: [])
    addresses : list = field(default_factory=lambda: [])


class KeyExtractedEvent(SniffingEvent):
    """Event indicating that a key has been extracted from pairing
    """

    def __init__(self, key):
        """Constructor."""
        super().__init__("Key extracted")
        self.key = key

    @property
    def message(self) -> str:
        """Show message."""
        return "key={}".format(self.key.hex())



from enum import IntEnum

class MACAddressMode(IntEnum):
    """
    Enum representing the different MAC address modes supported by 802.15.4.
    """
    NONE = 0
    SHORT = 1
    EXTENDED = 2

class MACDeviceType(IntEnum):
    """
    Enum representing the different MAC device types supported by 802.15.4.
    """
    RFD = 0
    FFD = 1

class MACPowerSource(IntEnum):
    """
    Enum representing the different MAC power source supported by 802.15.4.
    """
    BATTERY_SOURCE = 0
    ALTERNATING_CURRENT_SOURCE = 1

class MACScanType(IntEnum):
    """
    Enum representing MAC scan types supported by 802.15.4.
    """
    ENERGY_DETECTION = 0
    ACTIVE = 1
    PASSIVE = 2
    ORPHAN = 3
    ENHANCED_ACTIVE_SCAN = 4
    RIT_PASSIVE = 5

class MACNetworkType(IntEnum):
    """
    Enum representing the different network type supported by 802.15.4.
    """
    BEACON_ENABLED = 0
    NON_BEACON_ENABLED = 1


class MACBeaconType(IntEnum):
    """
    Enum representing the different beacon type supported by 802.15.4.
    """
    BEACON = 0
    ENHANCED_BEACON = 1

class MACAssociationStatus(IntEnum):
    """
    Enum representing the different association status supported by 802.15.4.
    """
    ASSOCIATION_SUCCESSFUL = 0
    PAN_AT_CAPACITY = 1
    PAN_ACCESS_DENIED = 2
    HOPPING_SEQUENCE_OFFSET_DUPLICATION = 3
    FAST_ASSOCIATION_SUCCESSFUL = 0x80

class MACConstants:
    """
    Constants used by 802.15.4 MAC layer.
    """
    A_BASE_SLOT_DURATION = 60
    A_NUM_SUPERFRAME_SLOTS = 16
    A_GTS_DESC_PERSISTENCE_TIME = 4
    A_MAX_LOST_BEACONS = 4
    A_MAX_SIFTS_FRAME_SIZE = 18
    A_MIN_CAP_LENGTH = 440
    A_BASE_SUPERFRAME_DURATION = 960
    A_UNIT_BACKOFF_PERIOD = None
    A_RCCN_BASE_SLOT_DURATION = 60

PANID_COMPRESSION_TABLE = {
    # (Dest. Addr, Src Addr, Dest. PAN ID, Src PAN ID) : pan_id_compress bit
    (MACAddressMode.NONE, MACAddressMode.NONE, False, False) : 0,

    (MACAddressMode.NONE, MACAddressMode.NONE, True, False) : 1,

    (MACAddressMode.SHORT, MACAddressMode.NONE, True, False) : 0,
    (MACAddressMode.EXTENDED, MACAddressMode.NONE, True, False) : 0,

    (MACAddressMode.SHORT, MACAddressMode.NONE, False, False) : 1,
    (MACAddressMode.EXTENDED, MACAddressMode.NONE, False, False) : 1,

    (MACAddressMode.NONE, MACAddressMode.SHORT, False, True) : 0,
    (MACAddressMode.NONE, MACAddressMode.EXTENDED, False, True) : 0,

    (MACAddressMode.NONE, MACAddressMode.SHORT, False, False) : 1,
    (MACAddressMode.NONE, MACAddressMode.EXTENDED, False, False) : 1,

    (MACAddressMode.EXTENDED, MACAddressMode.EXTENDED, True, False) : 0,

    (MACAddressMode.EXTENDED, MACAddressMode.EXTENDED, False, False) : 1,

    (MACAddressMode.SHORT, MACAddressMode.SHORT, True, True) : 0,
    (MACAddressMode.SHORT, MACAddressMode.EXTENDED, True, True) : 0,
    (MACAddressMode.EXTENDED, MACAddressMode.SHORT, True, True) : 0,

    (MACAddressMode.SHORT, MACAddressMode.EXTENDED, True, False) : 1,
    (MACAddressMode.EXTENDED, MACAddressMode.SHORT, True, False) : 1,
    (MACAddressMode.SHORT, MACAddressMode.SHORT, True, False) : 1,

}



"""
This module provides common converters for different data formats.
"""
from json import dumps
from dataclasses import fields

# Scapy
from scapy.packet import Packet
from scapy.fields import PacketListField, FlagValue

# WHAD
from whad.hub.metadata import Metadata

class ScapyConverter:
    """
    This class provides a basic API to convert a scapy packet to different
    representations and vice versa.
    """
    @classmethod
    def get_dict_from_scapy_packet(cls, pkt: Packet) -> dict:
        """
        This function converts a scapy packet to a python dictionary including
        the different layers and fields of the packet.

        :param pkt: packet to convert
        :type pkt: Packet
        :return: Packet converted into a dict
        :rtype: dict
        """

        pkt_dict = {}

        # Parsing of metadata structure
        if hasattr(pkt, "metadata"):
            pkt_dict["metadata"] = {}
            # We iterate over fields of the selected metadata class and
            # populate a sub dictionary linked to "metadata" key
            for field in fields(Metadata) + fields(pkt.metadata.__class__):
                try:
                    value = getattr(pkt.metadata, field.name)
                    # Add attributes if they are only defined
                    if value is not None:
                        if isinstance(value, (int, float)):
                            pkt_dict["metadata"][field.name] = value
                        else:
                            pkt_dict["metadata"][field.name] = str(value)
                except AttributeError:
                    pass

        # Iterate of each layer of the scapy packet
        for layer in pkt.layers():
            current_layer = {}
            # For a given layer, iterate over the fields of the scapy packet
            for field in layer.fields_desc:

                # In the case of a packet list field, call recursively the method to
                # process the sub-packets
                if isinstance(field, PacketListField) and field.cls is not None:
                    current_layer_list = []
                    for l in getattr(pkt[layer], field.name):
                        current_layer_list.append(cls.get_dict_from_scapy_packet(l))

                    # update the current layer with a list of sub-packets parsed as dict
                    current_layer[field.name] = current_layer_list
                else:
                    # General case, store the field name and its value in the dictionary
                    value = getattr(pkt[layer], field.name)

                    # Special processing for flags
                    if isinstance(value, FlagValue):
                        value = str(value)

                    # Special processing for bytes
                    elif isinstance(value,bytes):
                        value = value.hex()

                    current_layer[field.name] = value
            # update the dict with the current layer parsed as dict
            pkt_dict[layer.__name__] = current_layer

        # Return packet as a dict
        return pkt_dict

    @classmethod
    def get_json_from_scapy_packet(cls, pkt: Packet) -> str:
        """
        This function converts a scapy packet to JSON including the different
        layers and fields of the packet.

        :param pkt: packet to convert
        :type pkt: Packet
        """
        return dumps(cls.get_dict_from_scapy_packet(pkt))



"""Bluetooth Low Energy Central mock.

This mock emulates a WHAD device that supports BLE scanning and central modes,
emulating a set of devices with associated GATT servers.
"""

import logging
from time import time
from random import randint, choice
from queue import Queue, Empty
from typing import Union, Optional

from whad.exceptions import WhadDeviceDisconnected, WhadDeviceNotReady
from whad.hub import ProtocolHub
from whad.device.mock import MockDevice

from whad.hub.generic.cmdresult import Success, WrongMode, Error
from whad.hub.message import HubMessage
from whad.hub.ble.mode import CentralMode, BleStart, BleStop
from whad.hub.ble.pdu import BlePduReceived, SendBlePdu
from whad.hub.ble.connect import ConnectTo, Connected
from whad.hub.ble import BDAddress, Commands, Direction
from whad.hub.discovery import Capability, Domain, DeviceType

from .device import EmulatedDevice

logger = logging.getLogger(__name__)

class CentralMock(MockDevice):
    """BLE central mock device.

    This class implements the intended behavior of a BLE-compatible
    hardware interface.
    """

    def __init__(self, devices: Optional[list[EmulatedDevice]] = None, nowait: bool = False,
                 address: Optional[BDAddress] = None):
        """Initialize a mock BLE hardware interface that only allow device
        scanning / advertisement sniffing.

        :param devices: List of devices to emulate.
        :type devices: list[EmulatedDevice], optional
        :param nowait: Set to `True` to disable latency emulation
        :type nowait: bool, optional
        :param address: Bluetooth Device address associated with the emulated Central device. Default to 'aa:bb:cc:dd:ee:ff'
                        if not specified.
        :type address: whad.hub.ble.address.BDAddress, optional
        """
        if devices is None:
            logger.warning("No devices passed to scanner")
            self.__devices = []
        else:
            self.__devices = devices

        # Save nowait status
        self.__nowait = nowait

        # Status
        self.__central_mode_set = False
        self.__running = False
        self.__address = address or BDAddress("aa:bb:cc:dd:ee:ff", random=False)
        self.__conn_evt = None
        self.__conn_evt_ts = 0

        # Connections
        self.__target = None
        self.__conn_handle = -1
        self.__handles = [1, 17, 22, 34, 56]
        self.__connections = {}

        # Pending messages
        self.__messages = Queue()

        # Initialize our mock
        super().__init__(
            author="Whad Team",
            url="https://whad.io",
            proto_minver=ProtocolHub.LAST_VERSION,
            version="1.0.0",
            dev_type=DeviceType.VirtualDevice,
            dev_id=b"BleCentralMock",
            max_speed=115200,
            capabilities=self.__build_capabilities()
        )

    def __build_capabilities(self) -> dict:
        """Dynamically build the device's capabilities based on its config.
        """
        # By design, we don't capture raw BLE PDUs


        # Default commands
        commands = [
            Commands.CentralMode,
            Commands.Start, Commands.Stop,
            Commands.ConnectTo,
            Commands.SendPDU,
        ]
        capabilities = Capability.NoRawData

        return {
            Domain.BtLE : (
                # We can only scan and sniff with no raw data support
                capabilities,
                commands
            )
        }

    def on_interface_message(self):
        """Handle pending connection events.
        """
        # Handle pending connection event
        if self.__conn_evt is not None and self.__conn_evt_ts < time() and self.__target is not None:
            if self.__running:
                # Save connection event in connections
                self.__connections[self.__conn_evt.conn_handle] = self.__target

                # Set target device as connected
                self.__target.set_connected(self.__conn_handle)

                # Send notification
                self.put_message(self.__conn_evt)
                self.__conn_evt = None
                self.__conn_evt_ts = 0
                self.__conn_handle = -1
                self.__target = None

        try:
            msg = self.__messages.get(timeout=.5)
            return msg
        except Empty:
            return None

    @MockDevice.route(CentralMode)
    def on_central_mode(self, _: CentralMode):
        """BLE central mode handler"""
        if not self.__central_mode_set:
            if not self.__running:
                # Switch to scanning mode
                self.__central_mode_set = True
                return Success()
            else:
                # Return an error: another mode is running
                return WrongMode()

        if not self.__running:
            return Success()
        else:
            return Error()

    @MockDevice.route(BleStart)
    def on_start(self, _: BleStart):
        """Start selected mode"""
        if not self.__running:
            # If not running, mark as running and send a success response
            self.__running = True
            return Success()
        else:
            # If already running, return success
            return Success()

    @MockDevice.route(BleStop)
    def on_stop(self, _: BleStop):
        """Stop selected mode."""
        if self.__running:
            self.__running = False
        return Success()

    @MockDevice.route(ConnectTo)
    def on_connect(self, connect: ConnectTo):
        """Handle ConnectTo message.
        """
        # Check if the requested device belongs to our device list
        target = None
        for device in self.__devices:
            if device.address.value == connect.bd_address:
                # Device found
                target = device
                break

        # Generate a free connection handle, if possible
        if len(self.__handles) == 0:
            # No more connection handle available, return error
            return Error()
        else:
            conn_handle = choice(self.__handles)
            self.__handles.remove(conn_handle)

        # Device found ? "Connect" to this device and send a notification.
        if target is not None:
            connection_evt = Connected(
                access_address=0,
                initiator=self.__address.value,
                advertiser=target.address.value,
                conn_handle=conn_handle,
                adv_addr_type=target.addr_type,
                init_addr_type=self.__address.type,
            )

            if self.__nowait:
                # Save connection handle
                self.__connections[conn_handle] = target

                # Set target as connected
                target.set_connected(conn_handle)

                # Return success followed by a successful connected event.
                return [Success(), connection_evt]
            else:
                # Will generate a connection event later
                self.__target = target
                self.__conn_handle = conn_handle
                self.__conn_evt = connection_evt
                self.__conn_evt_ts = time()+randint(5, 20)/10.0
                return Success()

        # Won't find.
        self.__conn_evt = None
        self.__conn_evt_ts = 0
        return Success()

    @MockDevice.route(SendBlePdu)
    def on_send_pdu(self, send_pdu: SendBlePdu) -> Union[HubMessage, list[HubMessage]]:
        """Handle SendPdu command from connector.

        This callback forwards the given PDU to an existing connection.
        """
        if send_pdu.conn_handle in self.__connections:
            try:
                # Convert message to BLE_DATA object and send to our emulated device.
                answers = self.__connections[send_pdu.conn_handle].on_pdu(
                    send_pdu.to_packet()
                )

                result: list[HubMessage] = [Success()]

                # Convert response PDUs into BlePduReceived messages and add them
                # to the messages sent back to the connector
                for answer in answers:
                    result.append(BlePduReceived(
                        conn_handle=send_pdu.conn_handle,
                        direction=Direction.SLAVE_TO_MASTER,
                        pdu=bytes(answer),
                        processed=False,
                        decrypted=False
                    ))

                # Send messages to connector
                return result
            except (WhadDeviceDisconnected, WhadDeviceNotReady):
                return Error()
        else:
            return Error()




"""
APIMote adaptation layer.
"""
import select
import os
from struct import pack, unpack
from threading import Thread
from time import sleep

from serial import Serial,PARITY_NONE
from serial.tools.list_ports import comports

from scapy.layers.dot15d4 import Dot15d4FCS
from scapy.compat import raw


from whad.scapy.layers.apimote import GoodFET_Command_Hdr, GoodFET_Reply_Hdr, GoodFET_Init_Reply, \
    GoodFET_Peek_Command, GoodFET_Monitor_Connected_Command, GoodFET_Peek_Reply, \
    GoodFET_Setup_CCSPI_Command, GoodFET_Setup_CCSPI_Reply,GoodFET_Transfer_Command, \
    GoodFET_Transfer_Reply, GoodFET_Poke_Reply, GoodFET_Poke_Command, GoodFET_Peek_RAM_Command, \
    GoodFET_Poke_RAM_Command, GoodFET_Peek_RAM_Reply, GoodFET_Poke_RAM_Reply, \
    GoodFET_Read_RF_Packet_Command, GoodFET_Read_RF_Packet_Reply, GoodFET_Send_RF_Packet_Command, \
    GoodFET_Send_RF_Packet_Reply, CC_VERSIONS

from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady
from whad.hub.discovery import Domain, Capability
from whad.hub.generic.cmdresult import CommandResult
from whad.hub.dot15d4 import Commands

from whad.zigbee.utils.phy import channel_to_frequency, frequency_to_channel
from whad.device.uart import get_port_info

from ..device import VirtualDevice
from .constants import APIMoteId, APIMoteRegisters, \
    APIMoteRegistersMasks, APIMoteInternalStates


class Apimote(VirtualDevice):
    """Apimote virtual device implementation.
    """

    INTERFACE_NAME = "apimote"

    @classmethod
    def list(cls):
        '''
        Returns a list of available APIMote devices.
        '''
        available_devices = []
        for apimote in [
                        uart_dev for uart_dev in comports() if
                            (
                            uart_dev.vid == APIMoteId.APIMOTE_ID_VENDOR and
                            uart_dev.pid == APIMoteId.APIMOTE_ID_PRODUCT
                            )
                        ]:
            available_devices.append(Apimote(apimote.device))
        return available_devices

    @property
    def identifier(self):
        '''
        Returns the identifier of the current device (e.g., serial port).
        '''
        return self.__port


    def __init__(self, port, baudrate=115200):
        """
        Create device connection
        """
        super().__init__()

        # Connect to target UART device in non-blocking mode
        self.__port = port
        self.__baudrate = baudrate
        self.__fileno = None
        self.__uart = None
        self.__opened = False
        self.__synced = False
        self.__last_reply = None
        self.__input_data = None
        self._packet_queue = None
        self.__packet_polling = False
        self.__polling_thread = Thread(target=self._polling, daemon=True)

        self.__internal_state = APIMoteInternalStates.NONE
        port_info = get_port_info(self.__port)
        if port_info is None:
            raise WhadDeviceNotFound()

        self._dev_id = self._generate_dev_id(port_info)
        self._fw_author = self._get_author()
        self._fw_url = self._get_firmware_url()
        self._fw_version = self._get_firmware_version()
        self._dev_capabilities = self._get_capabilities()

    # Discovery related functions
    def _get_capabilities(self):
        capabilities = {
            Domain.Dot15d4 : (
                (Capability.Sniff | Capability.Inject),
                [Commands.Sniff, Commands.Send, Commands.Start, Commands.Stop]
            )
        }
        return capabilities

    def _generate_dev_id(self, port_info):
        dev_id = (
                        port_info.serial_number.encode('utf-8') +
                        pack("H", port_info.vid) +
                        pack("H", port_info.pid)
        )
        dev_id = b"\x00"*(16 - len(dev_id)) + dev_id
        return dev_id

    def _get_firmware_version(self):
        return (4, 0, 0) # APIMote v4 beta

    def _get_author(self):
        return "Ryan Speers".encode("utf-8")

    def _get_firmware_url(self):
        # Maybe we should replace it by GoodFET URL ?
        return "https://github.com/riverloopsec/apimote".encode("utf-8")

    def open(self):
        """
        Open device.
        """
        if not self.__opened:
            # Open UART device
            self.__uart = Serial(self.__port, self.__baudrate, parity = PARITY_NONE)


            # Get file number to use with select()
            self.__fileno = self.__uart.fileno()
            self.__opened = True

            self.__input_data = b""
            self.__uart.dtr = False             # Non reset state
            self.__uart.rts = False             # Non reset state
            self.__uart.dtr = self.__uart.dtr   # usbser.sys workaround

            # Ask parent class to run a background I/O thread
            super().open()

    def reset(self):
        """Reset device.

        This routine tries to reset device by setting RTS to high.
        """
        # If device is a true serial device, ask for a reset through DTR/RTS
        # Reset device through DTR
        self.__uart.dtr = False             # Non reset state
        self.__uart.rts = True             # Non reset state
        sleep(0.2)
        self.__uart.dtr = False             # Non reset state
        self.__uart.rts = False             # Non reset state

        while not self.__synced:
            sleep(0.1)
        self.__polling_thread.start()

        self._setup_ccspi()
        while self._peek_ccspi(APIMoteRegisters.MANFIDL) != 0x233D:
            sleep(0.1)

        self._init_radio()

    def read(self):
        """Read and process incoming data.
        """
        if not self.__opened:
            raise WhadDeviceNotReady()

        rlist = [self.__fileno]
        wlist = []
        elist = [self.__fileno]

        readers, _, __ = select.select(
            rlist,
            wlist,
            elist,
            1
        )

        # Handle incoming messages if any
        if len(readers) > 0:
            data = os.read(self.__fileno, 1024)
            self._process_input_data(data)


    def write(self, data: bytes) -> int:
        """Writes command to the device. It relies on select() in order to make sure
        we are allowed to write to the device and wait without eating too much CPU
        if the device is not ready to be written to.

        :param data: Data to write
        :type data: bytes
        :return: number of bytes written to the device
        :rtype: int
        """
        if not self.__opened:
            raise WhadDeviceNotReady()

        nb_bytes_written = 0
        wlist = [self.__fileno]
        elist = [self.__fileno]
        _ ,writers, __ = select.select(
            [],
            wlist,
            elist
        )

        if len(writers) > 0:
            nb_bytes_written = os.write(self.__fileno, raw(data))
        return nb_bytes_written


    def close(self):
        """
        Close current device.
        """
        # Ask parent class to stop I/O thread
        super().close()

        # Close underlying device.
        self.__opened = False
        self.__polling_thread.join()
        self.__uart.close()
        self.__uart = None
        self.__fileno = None

    def _send_whad_dot15d4_raw_pdu(self, packet, rssi=None):
        pdu = packet[:-2]
        fcs = unpack("H",packet[-2:])[0]
        is_fcs_valid = Dot15d4FCS().compute_fcs(pdu) == packet[-2:]
        timestamp = None

        msg = self.hub.dot15d4.create_raw_pdu_received(
            self._get_channel(),
            pdu,
            fcs,
            fcs_validity=is_fcs_valid
        )

        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp

        self._send_whad_message(msg)

    def _polling(self):
        while self.__opened:
            if self.__packet_polling:
                if self._packet_queue is None:
                    packet = self._get_packet()
                    if packet is not None:
                        rssi = self._get_rssi()
                        packet_size = packet[0]
                        # size of packet + 1-byte long length field
                        if packet_size + 1 < len(packet):
                            packet = packet[1:packet_size+1]
                            self._packet_queue = (packet[packet_size+2:], rssi)
                            self._send_whad_dot15d4_raw_pdu(packet, rssi)
                        else:
                            self._send_whad_dot15d4_raw_pdu(packet[1:], rssi)
                else:
                    packet = self._packet_queue
                    self._packet_queue = None
                    self._send_whad_dot15d4_raw_pdu(packet, rssi)

    # Virtual device whad message callbacks
    def _on_whad_dot15d4_stop(self, message):
        if self._switch_rf_to_idle():
            self.__internal_state = APIMoteInternalStates.NONE
            self.__packet_polling = False
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.ERROR)


    def _on_whad_dot15d4_send_raw(self, message):
        #print("here")
        channel = message.channel
        if self._set_channel(channel):
            packet = message.pdu
            old_state = self.__internal_state
            #self._switch_rf_to_tx()
            self.__internal_state = APIMoteInternalStates.TRANSMITTING
            self._configure_mdmctrl0(auto_crc=True)
            success = self._send_packet(packet)
            self._configure_mdmctrl0(auto_crc=False)


            if old_state == APIMoteInternalStates.SNIFFING:
                self._switch_rf_to_rx()
                self.__internal_state = APIMoteInternalStates.SNIFFING
            else:
                self._switch_rf_to_idle()
                self.__internal_state = APIMoteInternalStates.NONE

        else:
            success = False
        self._send_whad_command_result(CommandResult.SUCCESS if success else CommandResult.ERROR)


    def _on_whad_dot15d4_sniff(self, message):
        channel = message.channel
        if self._set_channel(channel):
            self.__internal_state = APIMoteInternalStates.SNIFFING
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

    def _on_whad_dot15d4_start(self, message):
        if self.__internal_state == APIMoteInternalStates.SNIFFING:
            if self._switch_rf_to_rx():
                self._packet_queue = None
                self.__packet_polling = True
                self._send_whad_command_result(CommandResult.SUCCESS)
            else:
                self._send_whad_command_result(CommandResult.ERROR)
        elif self.__internal_state == APIMoteInternalStates.NONE:
            self._switch_rf_to_idle()
            self._send_whad_command_result(CommandResult.SUCCESS)

    # APIMote / GoodFET low level primitives
    def _send_goodfet_cmd(self, cmd, app="MONITOR", reply_filter=None):
        #print("> cmd", repr(GoodFET_Command_Hdr(app=app)/cmd))
        self.write(raw(GoodFET_Command_Hdr(app=app)/cmd))
        if reply_filter is not None and callable(reply_filter):
            while self.__last_reply is None or not reply_filter(self.__last_reply):
                sleep(0.001)
            matched_reply = self.__last_reply
            self.__last_reply = None
            return matched_reply
        
        # Error
        return None

    def _process_goodfet_reply(self, reply):
        #print("< reply", repr(reply))

        if GoodFET_Init_Reply in reply and reply.url == b"http://goodfet.sf.net/":
            self.__synced = True
            self._monitor_connected()
        elif self.__synced:
            self.__last_reply = reply

    def _process_input_data(self, data):
        #print(data)

        self.__input_data += data
        if len(self.__input_data) >= 4:
            reply_length = (self.__input_data[2] | self.__input_data[3] << 8) + 4
            if len(self.__input_data) >= reply_length:
                self._process_goodfet_reply(GoodFET_Reply_Hdr(self.__input_data[:reply_length]))
                self.__input_data = self.__input_data[reply_length:]

    def _monitor_connected(self):
        self._send_goodfet_cmd(GoodFET_Monitor_Connected_Command())


    def _strobe_ccspi(self, register):
        data = bytes([register])
        status = self._transfer_ccspi(data)
        return status[0]


    def _poke_ccspi(self, register, value):
        data = bytes([
            register,
            0xFF & (value >> 8) ,
            (value & 0xFF)]
        )
        self._send_goodfet_cmd(
            GoodFET_Poke_Command(data=data),
            app="CCSPI",
            reply_filter=lambda reply:GoodFET_Poke_Reply in reply
        )
        return self._peek_ccspi(register) == value

    def _peek_ccspi(self, register):
        address = bytes([register, 0 , 0 ])
        reply = self._send_goodfet_cmd(
            GoodFET_Peek_Command(address=address),
            app="CCSPI",
            reply_filter=lambda reply:GoodFET_Peek_Reply in reply
        )
        value = 0
        for i in range(1,len(reply.data)):
            value |= (reply.data[i] << (len(reply.data)-i-1)*8)
        return value

    def _transfer_ccspi(self, data):
        reply = self._send_goodfet_cmd(
            GoodFET_Transfer_Command(data=data),
            app="CCSPI",
            reply_filter = lambda reply:GoodFET_Transfer_Reply in reply
        )
        return reply.data

    def _peek_ram(self, address, count=4):
        reply = self._send_goodfet_cmd(
                                        GoodFET_Peek_RAM_Command(address=address, count=count),
                                        app="CCSPI",
                                        reply_filter = lambda reply:GoodFET_Peek_RAM_Reply in reply
        )
        return reply.data

    def _poke_ram(self, address, value):
        data = bytes([address & 0xFF, (address & 0xFF00) >> 8]) + value
        self._send_goodfet_cmd(
            GoodFET_Poke_RAM_Command(data=data),
            app="CCSPI",
            reply_filter = lambda reply:GoodFET_Poke_RAM_Reply in reply
        )

    def _monitor_peek(self, address, size=8):
        if size == 8:
            reply = self._send_goodfet_cmd(
                GoodFET_Peek_Command(address=pack("H", address)),
                reply_filter=lambda reply:GoodFET_Peek_Reply in reply
            )
            return reply.data[0]

        if size == 16:
            reply_low = self._send_goodfet_cmd(
                GoodFET_Peek_Command(address=pack("H", address)),
                reply_filter=lambda reply:GoodFET_Peek_Reply in reply
            )
            reply_high = self._send_goodfet_cmd(
                GoodFET_Peek_Command(address=pack("H", address)+1),
                reply_filter=lambda reply:GoodFET_Peek_Reply in reply
            )
            return reply_low.data[0] + (reply_high.data[0] << 8)
        
        # Invalid size
        return None


    def _setup_ccspi(self):
        reply = self._send_goodfet_cmd(
            GoodFET_Setup_CCSPI_Command(),
            app="CCSPI",
            reply_filter = lambda reply:GoodFET_Setup_CCSPI_Reply in reply
        )
        return reply

    def _show_registers(self):
        '''
        Show current value of main registers (debugging purposes).
        '''
        print("MDMCTRL0", hex(self._peek_ccspi(APIMoteRegisters.MDMCTRL0)))
        print("MDMCTRL1", hex(self._peek_ccspi(APIMoteRegisters.MDMCTRL1)))
        print("IOCFG0", hex(self._peek_ccspi(APIMoteRegisters.IOCFG0)))
        print("SECCTRL0", hex(self._peek_ccspi(APIMoteRegisters.SECCTRL0)))

    # APIMote RF-related primitives
    def _init_radio(self):
        """
        Configure the radio as promiscuous and switch to idle.
        """
        self._setup_crystal_oscillator()
        self._setup_rf_calibration()
        self._configure_mdmctrl0(auto_ack=False, auto_crc=False, leading_zeroes=3,
                                 hardware_access_decoding=False, pan_coordinator=False,
                                 reserved_accepted=False)
        self._configure_mdmctrl1(demodulator_thresold=20)
        self._configure_iocfg0(filter_beacons=False)
        self._configure_secctrl0(enable_cbcmac=False, m=4, rx_key_select=0,
                                 tx_key_select=1, sa_key_select=1)
        self._switch_rf_to_idle()


    def _get_packet(self):
        """
        Return RX packet (if any is available).
        """
        reply = self._send_goodfet_cmd(
                                GoodFET_Read_RF_Packet_Command(),
                                app = "CCSPI",
                                reply_filter=lambda reply:True
        )
        if GoodFET_Read_RF_Packet_Reply in reply and reply.size > 0:
            return reply.data
        return None

    def _send_packet(self, packet):
        reply = self._send_goodfet_cmd(
            GoodFET_Send_RF_Packet_Command(data=bytes([len(packet)+2]) + packet),
            app = "CCSPI",
            reply_filter=lambda reply:True
        )
        return GoodFET_Send_RF_Packet_Reply in reply

    def _get_rssi(self):
        """
        Return last RSSI.
        """
        rssi = self._peek_ccspi(APIMoteRegisters.RSSI) & 0xFF
        # 2's complement, 8 bits
        if (rssi >> 7) == 1:
            rssi = rssi - (1 << 8)
        return rssi

    def _setup_crystal_oscillator(self):
        """
        Configure the crystal oscillator.
        """
        # Must be performed two times for some reason. See GOODFETCCSPI.py in Killerbee drivers.
        self._strobe_ccspi(APIMoteRegisters.SXOSCON)
        return self._strobe_ccspi(APIMoteRegisters.SXOSCON)

    def _setup_rf_calibration(self):
        """
        Setup the RF calibration.
        """
        return self._strobe_ccspi(APIMoteRegisters.STXCAL)

    def _switch_rf_to_idle(self):
        """
        Switch radio to idle mode.
        """
        return self._strobe_ccspi(APIMoteRegisters.SRFOFF)

    def _switch_rf_to_rx(self):
        """
        Switch radio to reception mode (RX).
        """
        return self._strobe_ccspi(APIMoteRegisters.SRXON)

    def _switch_rf_to_tx(self):
        """
        Switch radio to transmission mode (TX).
        """
        return self._strobe_ccspi(APIMoteRegisters.STXON)

    def _get_syncword(self):
        """
        Return syncword in use (default is 802.15.4 SFD).
        """
        return self._peek_ccspi(APIMoteRegisters.SYNCWORD)

    def _set_syncword(self, syncword = 0xA70F):
        """
        Configure syncword to use.
        """
        return self._poke_ccspi(APIMoteRegisters.SYNCWORD, syncword)

    def _get_channel(self):
        """
        Return channel currently in use.
        """
        return frequency_to_channel(self._get_frequency())

    def _set_channel(self, channel):
        """
        Configure channel to use.
        """
        if channel < 11 or channel > 26:
            return False
        self._set_frequency(channel_to_frequency(channel))
        return True

    def _get_frequency(self):
        """
        Return frequency in use.
        """
        masks = APIMoteRegistersMasks.FSCTRL
        fsctrl_value = self._peek_ccspi(APIMoteRegisters.FSCTRL)
        frequency_offset = (fsctrl_value & masks.FREQ.mask) >> masks.FREQ.offset
        return 2048+frequency_offset

    def _set_frequency(self, frequency):
        """
        Configure frequency to use.
        """
        masks = APIMoteRegistersMasks.FSCTRL
        fsctrl_value = self._peek_ccspi(APIMoteRegisters.FSCTRL)
        self._poke_ccspi( APIMoteRegisters.FSCTRL,
                            (fsctrl_value & ~(masks.FREQ.mask << masks.FREQ.offset)) |
                            ((int(frequency - 2048) & masks.FREQ.mask) << masks.FREQ.offset)
        )
        self._setup_rf_calibration()
        sleep(0.01)
        self._switch_rf_to_rx()

    def _configure_mdmctrl0(self, auto_ack=False, auto_crc=False, leading_zeroes=3,
                            hardware_access_decoding=False, pan_coordinator=False,
                            reserved_accepted=False):
        """
        Configure MDMCTRL0 register (manages various RF related features and hardware processing).
        """
        masks = APIMoteRegistersMasks.MDMCTRL0
        return self._poke_ccspi(APIMoteRegisters.MDMCTRL0,
            (
                ((int(reserved_accepted) & masks.RESERVED_FRAME_MODE.mask) << masks.RESERVED_FRAME_MODE.offset) |
                ((int(pan_coordinator) & masks.PAN_COORDINATOR.mask) << masks.PAN_COORDINATOR.offset) |
                ((int(hardware_access_decoding) & masks.ADR_DECODE.mask) << masks.ADR_DECODE.offset) |
                ((2 & masks.CCA_MODE.mask) << masks.CCA_HYST.offset) |
                ((3 & masks.CCA_MODE.mask) << masks.CCA_MODE.offset) |
                ((int(auto_ack) & masks.AUTO_ACK.mask) << masks.AUTO_ACK.offset) |
                ((int(auto_crc) & masks.AUTO_CRC.mask) << masks.AUTO_CRC.offset) |
                ((leading_zeroes-1 & masks.PREAMBLE_LENGTH.mask) << masks.PREAMBLE_LENGTH.offset)
            )
        )


    def _configure_mdmctrl1(self, demodulator_thresold=20):
        """
        Configure MDMCTRL1 register (manages various RF-related features).
        """
        masks = APIMoteRegistersMasks.MDMCTRL1
        return self._poke_ccspi(APIMoteRegisters.MDMCTRL1,
            (
                ((demodulator_thresold & masks.CORR_THR.mask) << masks.CORR_THR.offset) |
                ((0 & masks.DEMOD_AVG_MODE.mask) << masks.DEMOD_AVG_MODE.offset) |
                ((0 & masks.MODULATION_MODE.mask) << masks.MODULATION_MODE.offset) |
                ((0 & masks.TX_MODE.mask) << masks.TX_MODE.offset) |
                ((0 & masks.RX_MODE.mask) << masks.RX_MODE.offset)
            )
        )

    def _configure_iocfg0(self, filter_beacons=False):
        """
        Configure IOCFG0 register (manages polarity, beacon filtering and FIFO).
        """
        masks = APIMoteRegistersMasks.IOCFG0
        return self._poke_ccspi(APIMoteRegisters.IOCFG0,
            (
                ((int(not filter_beacons) & masks.BCN_ACCEPT.mask) << masks.BCN_ACCEPT.offset) |
                ((0 & masks.FIFO_POLARITY.mask) << masks.FIFO_POLARITY.offset) |
                ((0 & masks.FIFOP_POLARITY.mask) << masks.FIFOP_POLARITY.offset) |
                ((0 & masks.SFD_POLARITY.mask) << masks.SFD_POLARITY.offset) |
                ((0 & masks.CCA_POLARITY.mask) << masks.CCA_POLARITY.offset) |
                ((0x7F & masks.FIFOP_THR.mask) << masks.FIFOP_THR.offset)
            )
        )

    def _configure_secctrl0(self, enable_cbcmac=False, m=4, rx_key_select=0, tx_key_select=1,
                            sa_key_select=1):
        """
        Configure SECCTRL0 register (manages security-related features implemented in hardware).
        """
        masks = APIMoteRegistersMasks.SECCTRL0
        return self._poke_ccspi(APIMoteRegisters.SECCTRL0,
            (
                ((0 & masks.RXFIFO_PROTECTION.mask) << masks.RXFIFO_PROTECTION.offset) |
                ((1 & masks.SEC_CBC_HEAD.mask) << masks.SEC_CBC_HEAD.offset) |
                ((sa_key_select & masks.SEC_SAKEYSEL.mask) << masks.SEC_SAKEYSEL.offset) |
                ((tx_key_select & masks.SEC_TXKEYSEL.mask) << masks.SEC_TXKEYSEL.offset) |
                ((rx_key_select & masks.SEC_RXKEYSEL.mask) << masks.SEC_RXKEYSEL.offset) |
                ((int((m-2)//2) & masks.SEC_M.mask) << masks.SEC_M.offset) |
                ((int(enable_cbcmac) & masks.SEC_MODE.mask) << masks.SEC_MODE.offset)
            )
        )



"""wble-peripheral interactive shell.
"""
import json
from typing import Union, List, Tuple

# pylint: disable-next=wildcard-import,unused-wildcard-import
from scapy.layers.bluetooth4LE import *

from prompt_toolkit import print_formatted_text, HTML
from hexdump import hexdump

from whad.ble.exceptions import InvalidHandleValueException, \
    InvalidUUIDException as AttrInvalidUUIDException
from whad.ble.utils.validators import validate_attribute_uuid, InvalidUUIDException
from whad.exceptions import ExternalToolNotFound
from whad.device import WhadDevice, WhadDeviceConnector
from whad.ble import Peripheral, GenericProfile, AdvDataFieldList, \
    AdvCompleteLocalName, AdvShortenedLocalName, AdvFlagsField, AdvDataField, \
    AdvDataFieldListOverflow, AdvManufacturerSpecificData
from whad.ble.connector.peripheral import PeripheralEventListener, PeripheralEventConnected

from whad.ble.profile.service import PrimaryService
from whad.ble.profile.characteristic import Characteristic, CharacteristicProperties, \
    ClientCharacteristicConfig, CharacteristicValue
from whad.ble.profile.attribute import UUID
from whad.ble.stack.constants import BT_MANUFACTURERS
from whad.ble.stack.gatt.constants import CHARACS_UUID, SERVICES_UUID
from whad.ble.stack.att.exceptions import AttError, AttributeNotFoundError, \
    InsufficientAuthenticationError, InsufficientAuthorizationError, \
    InsufficientEncryptionKeySize, ReadNotPermittedError, \
    WriteNotPermittedError
from whad.common.monitors import WiresharkMonitor

from whad.cli.shell import InteractiveShell, category

INTRO='''
wble-periph, the WHAD Bluetooth Low Energy peripheral utility
'''

BDADDR_REGEXP = "^([a-fA-F0-9]{2}:){5}[a-fA-F0-9]{2}$"

ADRECORDS = {
}

def validate_uuid(uuid: str) -> str:
    """Ensure a user-provided UUID is valid.

    :param uuid: UUID to validate
    :type uuid: str
    :rtype: str
    :return: Validated UUID
    """
    try:
        # Parse provided UUID
        if isinstance(uuid, str) and uuid.lower().startswith("0x"):
            uuid = str(UUID(int(uuid, 16)))
        else:
            uuid = str(UUID(uuid))
    except TypeError as type_err:
        raise InvalidUUIDException(uuid) from type_err
    except InvalidUUIDException as uuid_err:
        raise InvalidUUIDException(uuid) from uuid_err

    # Return the normalized UUID
    return uuid

class AdvRecordsManager:
    """Advertising records manager

    This class manages the advertising data and scan response data of a
    device in order to optimize them.

    It is a bit limited for now as you may only have one field of each type
    in the advertising data, however it tries to store as much records as
    possible in the advertisement data and the scan response data.
    """

    def __init__(self, adv_data: AdvDataFieldList = None, scan_rsp_data: AdvDataFieldList = None):
        """Initialize manager
        """
        # Load advertising records
        self.__records = []
        if adv_data is not None:
            for r in adv_data:
                self.__records.append(r)
        if scan_rsp_data is not None:
            for r in scan_rsp_data:
                self.__records.append(r)

        # Save adv_data and scan_rsp_data in case no modification is
        # requested.
        self.__tainted = False
        self.__adv_data = adv_data
        self.__scan_rsp_data = scan_rsp_data

    @property
    def adv_data(self):
        """Return the advertising data (max 31 bytes)
        """
        # Pack records if they have been tainted
        if self.__tainted:
            self.pack()
        
        # Return the advertising data
        return self.__adv_data
    
    @property
    def scan_rsp_data(self):
        """Return the scan response data (max 31 bytes)
        """
        # Pack records if they have been tainted
        if self.__tainted:
            self.pack()

        # Return the scan response data
        return self.__scan_rsp_data

    @property
    def complete_name(self) -> str:
        """Parses current records and return the complete name value, if found.

        :return: Complete local name
        :rtype: str
        """
        cln = self.get_record(AdvCompleteLocalName)
        if cln is not None:
            return cln.name.decode("utf-8")

    @complete_name.setter
    def complete_name(self, value: str):
        """Update complete name
        """
        cln = self.get_record(AdvCompleteLocalName)
        if cln is not None:
            old_value = cln.name
            cln.name = bytes(value, "utf-8")
            try:
                self.update_record(cln)
            except AdvDataFieldListOverflow as overflow:
                cln.name = old_value
                self.update_record(cln)
                raise AdvDataFieldListOverflow() from overflow
        else:
            self.add_record(AdvCompleteLocalName(bytes(value, "utf-8")))

    @property
    def short_name(self) -> str:
        """Parses current records and return short name value, if found.

        :return: Short name
        :rtype: str
        """
        cln = self.get_record(AdvShortenedLocalName)
        if cln is not None:
            return cln.name.decode("utf-8")

    @short_name.setter
    def short_name(self, value: str):
        """Update shortened name
        """
        cln = self.get_record(AdvShortenedLocalName)
        if cln is not None:
            old_value = cln.name
            try:
                cln.name = bytes(value, "utf-8")
                self.update_record(cln)
            except AdvDataFieldListOverflow as overflow:
                cln.name = old_value
                self.update_record(cln)
                raise AdvDataFieldListOverflow() from overflow
        else:
            self.add_record(AdvShortenedLocalName(bytes(value, "utf-8")))

    @property
    def manufacturer_data(self):
        """Manufacturer data
        """
        cln = self.get_record(AdvManufacturerSpecificData)
        if cln is not None:
            return (cln.company, cln.data)
        
    @manufacturer_data.setter
    def manufacturer_data(self, data: Tuple[int, bytes]):
        if isinstance(data, tuple) and len(data) == 2:
            comp_id, data = data
            if isinstance(comp_id, int) and isinstance(data, bytes):
                cln = self.get_record(AdvManufacturerSpecificData)
                if cln is not None:
                    old_id, old_data = cln.company, cln.data
                    cln.company = comp_id
                    cln.data = data
                    try:
                        self.update_record(cln)
                    except AdvDataFieldListOverflow as overflow:
                        cln.company = old_id
                        cln.data = old_data
                        self.update_record(cln)
                        raise AdvDataFieldListOverflow() from overflow
                else:
                    self.add_record(AdvManufacturerSpecificData(comp_id, data))

    def __fit_records(self, records: List[AdvDataField], length: int = 31):
        """Find the records that may best fit in the given space
        """
        fitting_records = []

        # Sort records
        records.sort(key=lambda r: len(r.to_bytes()), reverse=True)
        
        # Find the biggest record that may fit in the given space
        used_space = 0
        found = True
        while found:
            found = False
            for r in records:
                if len(r.to_bytes()) <= length:
                    # Add record
                    fitting_records.append(r)
                    used_space += len(r.to_bytes())
                    length -= len(r.to_bytes())
                    records.remove(r)
                    found = True
                    break
        
        # Return the fitting records and the remaining list
        return (fitting_records, records)

    def pack(self):
        """Pack records into advertising and scan response data.
        """
        # Copy our records array
        records = [r for r in self.__records]

        # Pack records to fit advertising data
        adv_records, remaining = self.__fit_records(records, 31)

        # If some records remain, try to pack them into a scan response data
        if len(remaining) > 0:
            scan_rsp_records, remaining = self.__fit_records(records, 31)
        else:
            scan_rsp_records = None

        # If no remaining records, we're good.
        if len(remaining) > 0:
            raise AdvDataFieldListOverflow()
        else:
            # Update advertising data and scan
            self.__adv_data = AdvDataFieldList()
            for r in adv_records:
                self.__adv_data.add(r)
            if scan_rsp_records is not None:
                self.__scan_rsp_data = AdvDataFieldList()
                for r in scan_rsp_records:
                    self.__scan_rsp_data.add(r)
            
            # Advertising data and scan response data is up-to-date.
            self.__tainted = False

    def taint(self):
        """Mark advertising data as modified
        """
        if not self.__tainted:
            self.__tainted = True

    def get_record(self, record_type) -> AdvDataField:
        """Find a specific record.
        """
        if not self.__tainted:
            # Look into adv_data
            for r in self.__adv_data:
                if isinstance(r, record_type):
                    return r
                
            # Look into scan response data (if any)
            if self.__scan_rsp_data is not None:
                for r in self.__scan_rsp_data:
                    if isinstance(r, record_type):
                        return r
        else:
            # Look in our records
            for r in self.__records:
                if isinstance(r, record_type):
                    return r

        # No record found
        return None
    
    def remove_record(self, record_type) -> bool:
        """Remove a record of a specific type.
        """
        # Find and remove record
        record = self.get_record(record_type)
        if record is not None:
            # Remove record
            self.__records.remove(record)
            self.taint()

            # Success
            return True
        
        # Not found
        return False

    def update_record(self, record: AdvDataField):
        """Update a specific record.
        """
        # Remove record and add it again
        self.remove_record(record.__class__)
        self.add_record(record)
        
        # Make sure it fits
        self.pack()

    def add_record(self, record: AdvDataField) -> bool:
        """Add a record into the device advertising data, optimize storage.
        """
        # Mark as modified
        self.taint()

        # Add record to our records list
        self.__records.append(record)


class MonitoringProfile(GenericProfile):
    """Peripheral monitoring profile.
    """

    def on_characteristic_read(self, service, characteristic, offset=0, length=0):
        """Characteristic read hook.

        This hook is called whenever a characteristic is about to be read by a GATT client.
        If this method returns a byte array, this byte array will be sent back to the
        GATT client. If this method returns None, then the read operation will return an
        error (not allowed to read characteristic value).
        

        :param BlePrimaryService service: Service owning the characteristic
        :param BleCharacteristic characteristic: Characteristic object
        :param int offset: Read offset (default: 0)
        :param int length: Max read length
        :return: Value to return to the GATT client
        """
        print_formatted_text(HTML((
            f"<ansigreen>Reading</ansigreen> characteristic "
            f"<ansicyan>{characteristic.uuid}</ansicyan> of service "
            f"<ansicyan>{service.uuid}</ansicyan>"
        )))
        if len(characteristic.value) > 0:
            print_formatted_text(HTML(
                f" <i>{hexdump(characteristic.value, result='return')}</i>"
            ))
        else:
            print_formatted_text(HTML(" <i>Empty value</i>"))

    def on_connect(self, conn_handle):
        print_formatted_text(HTML(f"<ansired>New connection</ansired> handle:{conn_handle:d}"))

    def on_disconnect(self, conn_handle):
        print_formatted_text(HTML(f"<ansired>Disconnection</ansired> handle:{conn_handle:d}"))

    def on_characteristic_written(self, service, characteristic, offset=0, value=b'',
                                  without_response=False):
        """Characteristic written hook

        This hook is called whenever a charactertistic has been written by a GATT
        client.
        """
        print_formatted_text(HTML((
            f"<ansimagenta>Wrote</ansimagenta> to characteristic "
            f"<ansicyan>{characteristic.uuid}</ansicyan> of service "
            f"<ansicyan>{service.uuid}</ansicyan>"
        )))
        print_formatted_text(HTML(f" <i>{hexdump(value, result='return')}</i>"))


    def on_characteristic_subscribed(self, service, characteristic, notification=False,
                                     indication=False):
        # Check if we have a hook to call
        print_formatted_text(HTML((
            f"<ansicyan>Subscribed</ansicyan> to characteristic "
            f"<ansicyan>{characteristic.uuid}</ansicyan> of service "
            f"<ansicyan>{service.uuid}</ansicyan>"
        )))

    def on_characteristic_unsubscribed(self, service, characteristic):
        print_formatted_text(HTML((
            f"<ansicyan>Unsubscribed</ansicyan> to characteristic "
            f"<ansicyan>{characteristic.uuid}</ansicyan> of service "
            f"<ansicyan>{service.uuid}</ansicyan>"
        )))


class BlePeriphShell(InteractiveShell):
    """Bluetooth Low Energy interactive shell
    """

    MODE_NORMAL = 0
    MODE_SERVICE_EDIT = 1
    MODE_STARTED = 2

    def __init__(self, interface: WhadDevice = None, dev_profile=None):
        super().__init__(HTML("<b>wble-periph></b> "))

        self.__current_mode = self.MODE_NORMAL

        # Device parameters
        self.__complete_name = "WhadDev"
        self.__shortened_name = None
        self.__manuf_data = []
        self.__manuf_comp = None

        # If interface is None, pick the first matching our needs
        self.__interface = interface

        # Profile
        if dev_profile is not None:
            # Set profile
            self.__profile = MonitoringProfile(from_json=dev_profile)

            # Set advertising data
            profile = json.loads(dev_profile)
            if "devinfo" in profile and "adv_data" in profile["devinfo"]:
                self.adv_data = AdvDataFieldList.from_bytes(
                    bytes.fromhex(profile["devinfo"]["adv_data"])
                )
                if "scan_rsp" in profile["devinfo"]:
                    self.scan_rsp_data = AdvDataFieldList.from_bytes(
                        bytes.fromhex(profile["devinfo"]["scan_rsp"])
                    )
                else:
                    self.scan_rsp_data = None
            else:
                # No advertising data (should not be the case)
                self.warning("No advertising data in JSON profile !")
                self.adv_data = AdvDataFieldList(AdvFlagsField())
                self.scan_rsp_data = None
        else:
            # Create a default profile
            self.__profile = MonitoringProfile()

            # Generate AD data
            self.adv_data = AdvDataFieldList()
            if self.__complete_name is not None:
                self.adv_data.add(AdvCompleteLocalName(
                    bytes(self.__complete_name, "utf-8")
                ))
            if self.__shortened_name is not None:
                self.adv_data.add(AdvShortenedLocalName(
                    bytes(self.__shortened_name, "utf-8")
                ))
            self.adv_data.add(AdvFlagsField(
                bredr_support=False,
            ))

            # Set advertising data
            self.scan_rsp_data = None

        self.__adv_manager = AdvRecordsManager(self.adv_data, self.scan_rsp_data)
        self.__selected_service = None
        self.__connector: WhadDeviceConnector = None
        self.__listener: PeripheralEventListener = None
        self.__wireshark = None
        self.__central_bd = None
        self.intro = INTRO

        # Update prompt
        self.update_prompt()

    def update_prompt(self, force=False):
        """Update prompt to reflect current state
        """
        # Are we in service edit mode ?
        if self.__current_mode == self.MODE_SERVICE_EDIT:
            self.set_prompt(HTML(
                f"<b>wble-periph|<ansicyan>service({self.__selected_service})</ansicyan>></b> "),
                force)
        elif self.__current_mode == self.MODE_NORMAL:
            if not self.__central_bd:
                self.set_prompt(HTML("<b>wble-periph></b> "), force)
            else:
                self.set_prompt(
                    HTML(f"<b>wble-periph|<ansicyan>{self.__central_bd}</ansicyan>></b> "),
                    force)
        elif self.__current_mode == self.MODE_STARTED:
            self.set_prompt(
                HTML("<b>wble-periph<ansimagenta>[running]</ansimagenta>></b> "))


    def switch_role(self, new_role):
        """Switch from current role (if any) to another role.
        """
        if self.__connector is not None:
            self.__connector.stop()
            if self.__wireshark is not None:
                self.__wireshark.detach()
        self.__connector = new_role(self.__interface)
        if self.__wireshark is not None:
            self.__wireshark.attach(self.__connector)
            self.__wireshark.start()

    def show_att_error(self, error: AttError):
        """Parse ATT error and show exception.
        """
        if isinstance(error, InvalidHandleValueException):
            self.error("ATT Error: wrong value handle")
        elif isinstance(error, ReadNotPermittedError):
            self.error("ATT error: read operation not allowed")
        elif isinstance(error, WriteNotPermittedError):
            self.error("ATT error: write operation not allowed")
        elif isinstance(error, InsufficientAuthenticationError):
            self.error("ATT error: insufficient authentication")
        elif isinstance(error, InsufficientAuthorizationError):
            self.error("ATT error: insufficient authorization")
        elif isinstance(error, AttributeNotFoundError):
            self.error("ATT error: attribute not found")
        elif isinstance(error, InsufficientEncryptionKeySize):
            self.error("ATT error: insufficient encryption")

    ##################################################
    # GATT Service management
    ##################################################

    def has_service(self, uuid: str):
        """Check if a service is already registered
        """
        return self.__profile.get_service_by_uuid(UUID(uuid)) is not None

    def get_service(self, uuid: str):
        """Return service characteristics
        """
        service_obj = self.__profile.get_service_by_uuid(UUID(uuid))
        if service_obj is not None:
            return service_obj

        # Cannot find UUID
        raise IndexError

    def register_service(self, uuid: str):
        """Register a service

        @param  str     uuid    Service UUID
        @retval bool    True if service has been successfully registered, False otherwise
        """
        service_obj = self.__profile.get_service_by_uuid(UUID(uuid))
        if service_obj is None:
            self.__profile.add_service(PrimaryService(uuid=UUID(uuid)))
            return True

        return False

    def unregister_service(self, uuid: str):
        """Unregister a service
        """
        service_obj = self.__profile.get_service_by_uuid(UUID(uuid))
        if service_obj is not None:
            self.__profile.remove_service(service_obj)
            return True

        # Fail
        return False

    def enum_services(self):
        """Enumerate services
        """
        for service in self.__profile.services():
            yield (str(service.uuid), service)

    def select_service(self, uuid):
        """Select service and switch to edit mode.
        """
        self.__selected_service = uuid
        self.__current_mode = self.MODE_SERVICE_EDIT
        self.update_prompt()

    def complete_service(self):
        """auto-completion for 'service' command
        """
        services = [str(s.uuid) for s in list(self.__profile.services())]
        completions = {}
        for action in ["add", "edit", "remove"]:
            completions[action] = {}
            for service in services:
                completions[action][service]={}
        return completions

    @category("GATT profile")
    def do_service(self, args):
        """Manage peripheral's GATT services

        <ansicyan><b>service</b> [<i>ACTION</i> <i>[PARAMS, ...]</i>]</ansicyan>

        This command manages the registered services, with the following <i>ACTION</i>s:

        - <b>add</b>: add a service to the peripheral's GATT profile
        - <b>edit</b>: select a service for edition
        - <b>remove</b>: remove a service from the peripheral's GATT profile

        To add a service: <b>service</b> <i>add</i> <i>UUID</i>

        To edit a specific service: <b>service</b> <i>edit</i>

        To remove a service: <b>service</b> <i>remove</i> <i>UUID</i>

        By default, this command lists the registered services.
        """
        if self.__current_mode != self.MODE_NORMAL:
            if self.__current_mode == self.MODE_SERVICE_EDIT:
                self.error("Already editing services.")
            else:
                self.error("Cannot edit services while peripheral is running.")
            return

        if len(args) > 0:
            action = args[0].lower()
            if action == "add":
                if len(args) >= 2:
                    try:
                        # Validate UUID
                        service_uuid = validate_attribute_uuid(args[1])

                        # If service already exists, error.
                        if self.has_service(service_uuid):
                            self.error(f"Service {service_uuid} already exists !")
                            return

                        # Register service
                        self.register_service(service_uuid)

                        # Auto-select this service
                        self.select_service(service_uuid)

                        # Success
                        self.success(f"Service {service_uuid} successfully added.")
                    except InvalidUUIDException as bad_uuid:
                        self.error(bad_uuid.description)
                else:
                    self.error("You need to provide a valid UUID.")
            elif action == "remove":
                # Not allowed in edit mode
                if self.__current_mode == self.MODE_SERVICE_EDIT:
                    self.error((
                        "You cannot add or remove service in service edit mode. "
                        "Use <ansicyan>back</ansicyan> to exit edit mode and try again."))
                else:
                    if len(args) >= 2:
                        try:
                            # Validate UUID
                            service_uuid = validate_attribute_uuid(args[1])

                            if self.has_service(service_uuid):
                                self.unregister_service(service_uuid)
                                self.success(f"Successfully removed service {service_uuid}.")
                            else:
                                self.error(f"Service {service_uuid} is not a registered service.")
                        except InvalidUUIDException as bad_uuid:
                            self.error(bad_uuid.description)
                    else:
                        self.error("You need to provide a valid UUID.")
            elif action == "edit":
                if self.__current_mode == self.MODE_SERVICE_EDIT:
                    self.error("Already in edit mode.")
                    return

                if len(args) >= 2:
                    try:
                        # Validate UUID
                        service_uuid = validate_attribute_uuid(args[1])

                        if self.has_service(service_uuid):
                            self.select_service(service_uuid)
                            self.update_prompt()
                            return

                        # Error, not registered
                        self.error(f"Service {service_uuid} is not a registered service.")
                    except InvalidUUIDException as bad_uuid:
                        self.error(bad_uuid.description)
                else:
                    self.error("You need to provide a valid UUID.")
            else:
                self.error(f"Unknown action <i>{action}</i>.")
        else:
            # Enumerate services and store them in a list
            services = list(self.__profile.services())

            if len(services) > 0:
                for service in services:
                    # Resolve service name if 16-bit UUID
                    service_uuid = str(service.uuid)
                    if service.uuid.type == UUID.TYPE_16:
                        uuid_val = int(service_uuid, 16)
                        service_name = SERVICES_UUID.get(uuid_val)
                    else:
                        service_name = None

                    if service_name is not None:
                        print_formatted_text(HTML((
                            f"<ansicyan><b>Service {service.uuid}</b> ({service_name})</ansicyan>"
                            f" (handles from {service.handle:d} to {service.end_handle:d}):"
                        )))
                    else:
                        print_formatted_text(HTML((
                            f"<ansicyan><b>Service {service.uuid}</b></ansicyan>"
                            f" (handles from {service.handle:d} to {service.end_handle:d}):"
                        )))

                    characteristics = list(service.characteristics())
                    if len(characteristics) > 0:
                        for i, charac in enumerate(characteristics):

                            char_chevron = "├" if i < (len(characteristics) - 1) else "└"
                            handle_chevron = "│" if i < (len(characteristics) - 1) else " "


                            # Retrieve characteristic name if 16-bit UUID
                            charac_uuid = charac.uuid
                            charac_name = None
                            if charac_uuid.type == UUID.TYPE_16:
                                uuid_val = int(str(charac_uuid), 16)
                                charac_name = CHARACS_UUID.get(uuid_val)

                            properties = charac.properties
                            perms = []
                            if properties & CharacteristicProperties.READ != 0:
                                perms.append("read")
                            if properties & CharacteristicProperties.WRITE != 0:
                                perms.append("write")
                            if properties & CharacteristicProperties.INDICATE != 0:
                                perms.append("indicate")
                            if properties & CharacteristicProperties.NOTIFY != 0:
                                perms.append("notify")

                            if charac_name is not None:
                                print_formatted_text(HTML((
                                    f"{char_chevron}─ <ansicyan><b>Characteristic "
                                    f"{charac.uuid}</b> ({charac_name})</ansicyan>"
                                )))
                            else:
                                print_formatted_text(HTML((
                                    f"{char_chevron}─ <ansicyan><b>Characteristic "
                                    f"{charac.uuid}</b></ansicyan>"
                                )))
                            print_formatted_text(HTML((
                                f"{handle_chevron} └─ handle:{charac.handle:d}, "
                                f"value handle: {charac.value_handle:d}, props: {','.join(perms)}"
                            )))

                            for desc in charac.descriptors():
                                print_formatted_text(HTML((
                                    f"{handle_chevron} └─ <b>Descriptor {desc.type_uuid}</b>"
                                    f" (handle: {desc.handle:d})"
                                )))
                    else:
                        print_formatted_text(HTML(" <i>No characteristics defined</i>"))
            else:
                self.error("No service registered yet.")

    ##################################################
    # GATT Characteristic management
    ##################################################

    def has_service_char(self, service_uuid:str, char_uuid:str):
        """Check if a characteristic is already registered for a specific service
        """
        service_obj = self.__profile.get_service_by_uuid(UUID(service_uuid))
        if service_obj is not None:
            charac = service_obj.get_characteristic(UUID(char_uuid))
            return charac is not None

        # Not found
        return False


    def register_char(self, service_uuid: str, char_uuid: str, perms: list):
        """Add a characteristic

        @param  service_uuid    Service UUID
        @param  char_uuid       Characteristic UUID
        @param  perms           List of permissions
        @retval                 True on success, False on failure
        """
        if not self.has_service_char(service_uuid, char_uuid):
            # Get service
            service_obj = self.get_service(service_uuid)

            # Build charac properties
            props = 0
            if "read" in perms:
                props |= CharacteristicProperties.READ
            if "write" in perms:
                props |= CharacteristicProperties.WRITE
            if "writecmd" in perms:
                props |= CharacteristicProperties.WRITE_WITHOUT_RESPONSE
            if "notify" in perms:
                props |= CharacteristicProperties.NOTIFY
            if "indicate" in perms:
                props |= CharacteristicProperties.INDICATE

            # Build characteristic object
            charac_obj = Characteristic(
                uuid=UUID(char_uuid),
                properties=props,
            )

            if "notify" in perms or "indicate" in perms:
                cccd = ClientCharacteristicConfig(
                    characteristic=charac_obj,
                    indicate=("indicate" in perms),
                    notify=("notify" in perms)
                )

                charac_obj.add_descriptor(cccd)

                charac_obj.value = b''

            # Add characteristic to service
            service_obj.add_characteristic(charac_obj)

            # Update service in profile
            self.__profile.update_service(service_obj)

            # Success
            self.success(f"Successfully added characteristic {char_uuid}")
            return True

        # Fail
        return False

    def unregister_char(self, service_uuid: str, char_uuid: str):
        """Remove characteristic from service
        """
        if self.has_service_char(service_uuid, char_uuid):
            # Get service
            service_obj = self.get_service(service_uuid)

            charac = service_obj.get_characteristic(UUID(char_uuid))
            if charac is not None:
                service_obj.remove_characteristic(charac)
                self.__profile.update_service(service_obj)
                return True

        # Fail
        return False

    def char_parse_perms(self, permissions):
        """Parse permissions
        """
        allowed_keywords = [
            "read",
            "write",
            "writecmd",
            "notify",
            "indicate"
        ]

        out_perms = []
        for perm in permissions:
            if perm.lower() in allowed_keywords:
                out_perms.append(perm.lower())
        return out_perms


    def complete_char(self):
        """Auto-complete char command
        """
        completions = {}
        if self.__selected_service is not None:
            service = self.__profile.get_service_by_uuid(UUID(self.__selected_service))
            if service is not None:
                chars = list(service.characteristics())
                for action in ["add", "remove"]:
                    completions[action] = {}
                    for char in chars:
                        completions[action][str(char.uuid)] = {}
        return completions

    @category("GATT profile")
    def do_char(self, args):
        """Manage peripheral's GATT characteristics

        <ansicyan><b>char</b> [<i>ACTION</i> <i>[PARAMS]</i>]</ansicyan>

        This command manages the registered characteristics of the currently
        selected service. The following <i>ACTION</i>s are available:

        - <b>add</b>: add a characteristic to the current service
        - <b>remove</b>: remove a characteristic from the current service

        charac add 2a00 read write notify
        """
        # Characteristics must be modified only in edit mode
        if self.__current_mode == self.MODE_SERVICE_EDIT:
            if len(args)>0:
                action = args[0]
                if action == "add":
                    if len(args)>=2:
                        try:
                            # Retrieve characteristic UUID
                            char_uuid = validate_attribute_uuid(args[1])

                            # Parse permissions
                            if len(args) >= 3:
                                perms = self.char_parse_perms(args[2:])
                            else:
                                perms = ["read"]

                            if not self.has_service_char(self.__selected_service, char_uuid):
                                self.register_char(self.__selected_service, char_uuid, perms)
                            else:
                                self.error((f"Characteristic {char_uuid} already exist in "
                                           f"service {self.__selected_service}"))
                        except InvalidUUIDException as bad_uuid:
                            self.error(bad_uuid.description)
                    else:
                        print_formatted_text(HTML(
                            "<b>Usage:</b> <ansicyan>add</ansicyan> <i>UUID</i> [<i>PERM,</i> ...]"
                        ))
                elif action == "remove":
                    if len(args)>=2:
                        try:
                            # Retrieve characteristic UUID
                            char_uuid = validate_attribute_uuid(args[1])

                            # Remove characteristic
                            if self.has_service_char(self.__selected_service, char_uuid):
                                self.unregister_char(self.__selected_service, char_uuid)
                                self.success(f"Successfully removed characteristic {char_uuid}")
                            else:
                                self.error(f"Characteristic {char_uuid} does not exist.")
                        except InvalidUUIDException as bad_uuid:
                            self.error(bad_uuid.description)
                    else:
                        print_formatted_text(HTML(
                            "<b>Usage:</b> <ansicyan>remove</ansicyan> <i>UUID</i>"
                        ))
            else:
                # List characteristics
                print_formatted_text(HTML(
                    f"<ansicyan>Characteristics for service {self.__selected_service}:</ansicyan>"
                ))
                selected_service = self.__profile.get_service_by_uuid(UUID(self.__selected_service))
                if selected_service is not None:
                    characs = list(selected_service.characteristics())
                    if len(characs) > 0:
                        for charac in characs:
                            uuid = charac.uuid
                            if uuid.type == UUID.TYPE_16:
                                uuid_val = int(str(uuid), 16)
                                charac_name = CHARACS_UUID.get(uuid_val)
                            else:
                                charac_name = None

                            properties = charac.properties
                            perms = []
                            if properties & CharacteristicProperties.READ != 0:
                                perms.append("read")
                            if properties & CharacteristicProperties.WRITE != 0:
                                perms.append("write")
                            if properties & CharacteristicProperties.INDICATE != 0:
                                perms.append("indicate")
                            if properties & CharacteristicProperties.NOTIFY != 0:
                                perms.append("notify")

                            access = ",".join([f"<b>{perm}</b>" % perm for perm in perms])
                            if charac_name is not None:
                                print_formatted_text(HTML(
                                    f" {charac.uuid} ({charac_name}): {access}"
                                ))
                            else:
                                print_formatted_text(HTML(
                                    f" {charac.uuid}: ({access})"
                                ))
                    else:
                        print_formatted_text(HTML(" No characteristic registered."))


    def char_set_value(self, handle_uuid, value):
        """Set characteristic value based on its handle or UUID
        """
        if isinstance(handle_uuid, int):
            if self.__current_mode != self.MODE_STARTED:
                self.error("Cannot set a characteristic value by its handle")

            # Search characteristic by its handle
            if self.__profile is not None:
                try:
                    charac = self.__profile.find_object_by_handle(handle_uuid)
                    if isinstance(charac, Characteristic):
                        # Update value
                        charac.value = value
                except IndexError:
                    self.error("Cannot find characteristic with UUID %s" % handle_uuid)
            else:
                self.error("GATT profile has not been set.")
        elif isinstance(handle_uuid, UUID):
            # Are we started ?
            if self.__current_mode == self.MODE_STARTED:
                # Find characteristic handle
                for _, service in self.enum_services():
                    charac_obj = service.get_characteristic(handle_uuid)
                    if charac_obj is not None:
                        charac_obj.value = value
                        return
            else:
                # Find characteristic handle
                for _, service in self.enum_services():
                    charac_obj = service.get_characteristic(handle_uuid)
                    if charac_obj is not None:
                        charac_obj.value = value

    def perform_write(self, args, without_response=False):
        """Perform attribute/handle characteristic
        """
        # parse target arguments
        if len(args) <2:
            self.error(("You must provide at least a characteristic value handle"
                        " or characteristic UUID, and a value to write."))
            return

        handle = None

        # Figure out what the handle is
        if args[0].lower().startswith("0x"):
            try:
                handle = int(args[0].lower(), 16)
            except ValueError:
                self.error(f"Wrong handle: {args[0]}")
                return
        else:
            try:
                handle = int(args[0])
            except ValueError:
                try:
                    handle = UUID(args[0].replace("-",''))
                except Exception:
                    self.error(f"Wrong UUID: {args[0]}")
                    return

        # Do we have hex data ?
        if args[1].lower() == "hex":
            # Decode hex data
            hex_data = ''.join(args[2:])
            try:
                char_value = bytes.fromhex(hex_data.replace("\t", ""))
            except ValueError:
                self.error("Provided hex value contains non-hex characters.")
                return
        else:
            char_value = args[1]

        if not isinstance(char_value, bytes):
            char_value = bytes(char_value,"utf-8")

        # Update characteristic value
        self.char_set_value(handle, char_value)

    @category("GATT profile")
    def do_writecmd(self, args):
        """write data to a GATT attribute without waiting for a response.

        <ansicyan><b>writecmd</b> <i>[UUID | handle] [hex [value] | value ]</i></ansicyan>

        Write data to the specified GATT attribute (identified by its handle) or to
        a characteristic value (identified by its UUID, if unique) without waiting
        for a response.

        Data can be provided hex-encoded if prefixed by "hex":

        > writecmd 41 hex 41 42 43

        The command above will write 'ABC' to attribute identified by the handle 41.
        
        Data can also be provided as text:

        > writecmd 41 ABC
        """
        self.perform_write(args, without_response=True)

    @category("GATT profile")
    def do_write(self, args):
        """write data to a a GATT attribute.

        <ansicyan><b>write</b> <i>[UUID | handle] [hex [value] | value ]</i></ansicyan>

        Write data to the specified GATT attribute (identified by its handle) or to
        a characteristic value (identified by its UUID, if unique).

        Data can be provided hex-encoded if prefixed by "hex":

        > write 41 hex 41 42 43

        The command above will write 'ABC' to attribute identified by the handle 41.
        
        Data can also be provided as text:

        > write 41 ABC

        """
        self.perform_write(args, without_response=False)

    @category("GATT profile")
    def do_read(self, args):
        """read a GATT attribute

        <ansicyan><b>read</b> <i>[UUID | handle]</i></ansicyan>

        Read an attribute identified by its handle, or read the value of a characteristic
        identified by its UUID (if unique). 

        Result is displayed as an hexadecimal dump with corresponding ASCII text:

        > read 41
         00000000: 74 68 69 73 20 69 73 20  61 20 74 65 73 74        this is a test
        """
        if len(args) >= 1:
            try:
                charac_uuid = UUID(args[0])
                charac = self.__profile.get_characteristic_by_uuid(charac_uuid)
                if charac is not None:
                    print_formatted_text(HTML(
                        f" <i>{hexdump(charac.value, result='return')}</i>"
                    ))
                else:
                    self.error("Unknown characteristic {charac_uuid}")
            except AttrInvalidUUIDException:
                try:
                    # Decode handle
                    if args[0].lower().startswith("0x"):
                        charac_handle = int(args[0], 16)
                    else:
                        charac_handle = int(args[0], 10)

                    # Retrieve characteristic by handle
                    charac = self.__profile.find_object_by_handle(charac_handle)
                    if isinstance(charac, Characteristic):
                        print_formatted_text(HTML(
                            f" <i>{hexdump(charac.value, result='return')}</i>"
                        ))
                    elif isinstance(charac, CharacteristicValue):
                        print_formatted_text(HTML(
                            f" <i>{hexdump(charac.characteristic.value, result='return')}</i>"
                        ))
                    else:
                        self.error(f"Unknown characteristic with handle {charac_handle:d}")
                except ValueError:
                    self.error(f"Wrong UUID or handle: {args[0]}")


    ##################################################
    # Peripheral emulation
    ##################################################

    @category("Peripheral control")
    def do_start(self, _):
        """Start peripheral.

        <ansicyan><b>start</b></ansicyan>

        Starts the peripheral with its configured services and characteristics.
        """

        # Switch to emulation mode
        self.__current_mode = self.MODE_STARTED
        self.update_prompt()

        try:
            # Instantiate our Peripheral
            self.__connector = Peripheral(
                self.__interface,
                profile=self.__profile,
                adv_data=self.__adv_manager.adv_data,
                scan_data=self.__adv_manager.scan_rsp_data
            )

            # Start advertising
            self.__connector.start()
        except AdvDataFieldListOverflow:
            self.error("Advertising data is too big to fit in advertisement !")
        
        # Instantiate our Peripheral
        # self.__connector = Peripheral(
        #    self.__interface,
        #    profile=self.__profile,
        #    adv_data=adv_data
        #)
        # create our event listener
        # self.__listener = PeripheralEventListener(callback=self.on_periph_event)
        # self.__listener.start()
        # self.__connector.attach_event_listener(self.__listener)

        # Start peripheral
        # self.__connector.start()


    def on_periph_event(self, event):
        if isinstance(event, PeripheralEventConnected):
            print("Got a connection from a central, updating MTU")
            self.__connector.set_mtu(200)

    @category("Peripheral control")
    def do_stop(self, _):
        """Stop peripheral
        """
        if self.__connector is not None:
            self.__connector.stop()
            self.__connector.close()
        
        if self.__listener is not None:
            self.__listener.stop()
            self.__listener.join()

        self.__current_mode = self.MODE_NORMAL
        self.update_prompt()


    def complete_wireshark(self):
        """Autocomplete wireshark command
        """
        completions = {}
        if self.__wireshark is not None:
            completions["off"] = {}
        else:
            completions["on"] = {}
        return completions

    @category("Monitoring")
    def do_wireshark(self, arg):
        """launch wireshark to monitor packets

        <ansicyan><b>wireshark</b> <i>["on" | "off"]</i></ansicyan>

        This command launches a wireshark that will display all the packets sent
        and received in the active connection. 
        """
        if len(arg) >=1:
            enabled = arg[0].lower()=="on"
            if enabled:
                if self.__wireshark is None:
                    try:
                        self.__wireshark = WiresharkMonitor()
                        if self.__connector is not None:
                            self.__wireshark.attach(self.__connector)
                            self.__wireshark.start()
                    except ExternalToolNotFound:
                        self.error("Cannot launch Wireshark, please make sure it is installed.")
                else:
                    self.error(("Wireshark is already launched, see "
                                "<ansicyan>wireshark off</ansicyan>"))
            else:
                # Detach monitor if any
                if self.__wireshark is not None:
                    self.__wireshark.detach()
                    self.__wireshark.close()
                    self.__wireshark = None
        else:
            self.error("Missing arguments, see <ansicyan>help wireshark</ansicyan>.")

    @category("Advertising data")
    def do_name(self, args):
        """Set device advertising complete local name

        <ansicyan><b>name</b> <i>DEVICE NAME</i></ansicyan>

        This command sets the complete local name for the emulated peripheral.
        """
        if len(args) > 0:
            # Save name
            name = args[0]

            try:
                self.__adv_manager.complete_name = name
                self.success(f"Device name set to \"{name}\"")
            except AdvDataFieldListOverflow:
                self.error("Advertising data is full, cannot set complete local name.")

        else:
            # Complete name
            complete_name = self.__adv_manager.complete_name
            if complete_name is not None:
                print_formatted_text(HTML(
                    f"<ansicyan>Device complete name:</ansicyan> {complete_name}"
                ))
            else:
                self.warning("Complete local name not set.")

    @category("Advertising data")
    def do_shortname(self, args):
        """Set device short name

        <ansicyan><b>shortname</b> <i>DEVICE NAME</i></ansicyan>

        This command sets the shortened local name for the emulated peripheral.
        """
        if len(args) > 0:
            # Save name
            name = args[0]

            try:
                self.__adv_manager.short_name = name
                self.success(f"Device shortened local name set to \"{name}\"")
            except AdvDataFieldListOverflow:
                self.error("Advertising data is full, cannot set shortened local name")
        else:
            # Short name
            short_name = self.__adv_manager.short_name
            if short_name is not None:
                print_formatted_text(HTML(
                    f"<ansicyan>Device shortened name:</ansicyan> {short_name}"
                ))
            else:
                self.error("No shortened local name has been set yet.")

    @category("Advertising data")
    def do_manuf(self, args):
        """Set device manufacturer company ID and data.

        <ansicyan><b>manuf</b> [<i>COMPANY_ID</i> <i>HEX_DATA</i>]</ansicyan>

        This commands defines the company ID to use in BLE advertising record
        along with manufacturer data if a COMPANY_ID and HEX_DATA are provided,
        or simply manufacturer data otherwise.
        """
        if len(args) >= 2:
            comp_id = args[0]
            manuf_data = args[1]
            manuf_comp = None
            if comp_id.lower().startswith("0x"):
                manuf_comp = int(comp_id, 16)
            else:
                # First, search for company in our list of known companies (no case compare)
                for cid, comp_name in BT_MANUFACTURERS.items():
                    if comp_name.lower() == comp_id:
                        manuf_comp = cid
                        break

                if manuf_comp is None:
                    try:
                        manuf_comp = int(comp_id)
                    except ValueError:
                        self.error("Bad company ID ({comp_id})")
                        return

            try:
                self.__adv_manager.manufacturer_data = (manuf_comp, bytes.fromhex(manuf_data))
                self.success("Manufacturer data set.")
            except ValueError:
                self.__manuf_data = []
                self.error("Error while parsing manufacturer data (not valid hex)")
            except AdvDataFieldListOverflow:
                self.error("Advertising data is full, cannot set manufacturer specific data")
        else:
            if self.__adv_manager.manufacturer_data is not None:
                manuf_comp, manuf_data = self.__adv_manager.manufacturer_data
                if manuf_comp in BT_MANUFACTURERS:
                    manuf_name = BT_MANUFACTURERS[manuf_comp]
                else:
                    manuf_name = "Unknown"
                print_formatted_text(HTML("<ansicyan>Device Manufacturer data record:</ansicyan>"))
                print_formatted_text(HTML(
                    f" <b>Company ID:</b> 0x{manuf_comp:04x} ({manuf_name})"
                ))
                if manuf_data != []:
                    print_formatted_text(HTML(" <b>Manuf. Data:</b>"))
                    nb_lines = int(len(manuf_data)/16)
                    if nb_lines*16 < len(manuf_data):
                        nb_lines += 1
                    for i in range(nb_lines):
                        data = " ".join(
                            [f"{v:02x}" for v in manuf_data[i*16:(i+1)*16]]
                        )
                        print_formatted_text(HTML(f"   <i>{data}</i>"))
            else:
                self.error("No manufacturer data has been set yet.")

    @category("Peripheral control")
    def do_mtu(self, args):
        """Set peripheral MTU.

        <ansicyan><b>mtu</b> [<i>MTU</i>]</ansicyan>

        Send a MTU exchange request to the connected Central device.
        MTU value must be equal to or greater than 23.
        """
        if len(args) == 1:
            try:
                mtu = int(args[0])

                # Make sure we have a connector
                if self.__connector is None:
                    self.error("No active connection, cannot set MTU.")
                    return

                # Update MTU value
                if mtu >= 23:
                    self.__connector.set_mtu(mtu)
                    print(f"Connection MTU set to {mtu}.")
                else:
                    self.error("MTU must be greater or equal to 23.")
            except ValueError:
                self.error("MTU is not a valid integer")
        elif len(args) < 1:
            self.error("MTU value is missing")
        elif len(args) > 1:
            self.error("Too many arguments !")

    def do_back(self, _):
        """Return to normal mode.

        <ansicyan><b>back</b></ansicyan>

        This command return to normal mode, i.e. exits any configuration mode
        (service edit mode for instance).
        """
        if self.__current_mode != self.MODE_NORMAL:
            self.__current_mode = self.MODE_NORMAL
            self.__selected_service = None
        self.update_prompt()

    def do_quit(self, _):
        """close ble-peripheral
        """
        if self.__connector is not None:
            self.__connector.close()
        return True

    def do_exit(self, arg):
        """alias for <ansicyan>quit</ansicyan>
        """
        return self.do_quit(arg)



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
import sys
import logging
from time import sleep

from scapy.config import conf
from scapy.themes import BrightTheme

from whad.cli.ui import wait, success
from whad.hub import ProtocolHub
from whad.device import Bridge
from whad.device.connector import WhadDeviceConnector
from whad.device.unix import UnixConnector, UnixSocketServer
from whad.common.monitors import WiresharkMonitor
from whad.cli.app import CommandLineApp, run_app

# wshark logger
logger = logging.getLogger(__name__)

class WhadWiresharkApp(CommandLineApp):
    """Main WiresharkApp class
    """

    # App connector
    connector = None

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD wireshark',
            interface=True,
            commands=False,
            input=CommandLineApp.INPUT_WHAD,
            output=CommandLineApp.OUTPUT_WHAD
        )

        # Initialize parameters.
        self.monitor = None

    def run(self):
        """Application main() routine
        """

        # Launch pre-run tasks
        self.monitor = None
        self.pre_run()
        try:
            if self.is_piped_interface():
                interface = self.input_interface
            else:
                interface = self.interface

            if interface is not None:
                if not self.args.nocolor:
                    conf.color_theme = BrightTheme()

                # Using a UnixConnector is just a small hack, as it acts like
                # a dummy connector.
                connector = UnixConnector(interface)
                connector.domain = self.args.domain
                hub = ProtocolHub(2)
                connector.format = hub.get(self.args.domain).format

                # Attack a wireshark monitor
                self.monitor = WiresharkMonitor()
                self.monitor.attach(connector)
                self.monitor.start()

                if self.is_stdout_piped():
                    proxy = UnixConnector(UnixSocketServer(parameters=self.args.__dict__))

                    while not proxy.device.opened:
                        if proxy.device.timedout:
                            return
                        sleep(0.1)

                    # Bridge both connectors and their respective interfaces
                    _ = Bridge(connector, proxy)

                    # Wait for the user to CTL-C or close Wireshark
                    while interface.opened and not self.monitor.is_terminated():
                        sleep(.1)

                else:
                    connector.unlock()

                    # Wait for the user to CTL-C or close Wireshark
                    while interface.opened and not self.monitor.is_terminated():
                        wait(f"Forwarding {self.monitor.packets_written} packets to wireshark")
                        sleep(.2)

        except KeyboardInterrupt:
            # Launch post-run tasks
            if self.monitor is not None:
                self.monitor.stop()
                self.monitor.close()
        finally:
            self.post_run()

    def post_run(self):
        if not self.is_stdout_piped() and self.monitor is not None:
            wait(f"Forwarding {self.monitor.packets_written} packets to wireshark",
                end=True
            )
            success(f"{self.monitor.packets_written} packets have been forwarded to wireshark")
        super().post_run()


def wshark_main():
    """Launcher for wshark.
    """
    app = WhadWiresharkApp()
    run_app(app)



from whad.unifying.connector import Unifying, Mouse
from whad.esb.stack import ESBStack
from whad.unifying.stack import UnifyingApplicativeLayer
from whad.exceptions import UnsupportedCapability, WhadDeviceDisconnected
from whad.unifying.injecting import InjectionConfiguration
from whad.helpers import message_filter, is_message_type
from time import sleep, time


class Injector(Mouse):
    """
    Logitech Unifying Injector interface.
    """
    def __init__(self, device):
        super().__init__(device)
        self._autosync = True
        self._synced = False
        self.__configuration = InjectionConfiguration()
        self._injecting = False
        self.start()

    @property
    def configuration(self) -> InjectionConfiguration:
        """Retrieve this injector configuration.
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration: InjectionConfiguration):
        """Set the injector configuration.
        """
        self.__configuration = new_configuration
        if self.__configuration.address is not None:
            self.address = self.__configuration.address

    def on_pdu(self, pdu):
        if self._injecting:
            return
        else:
            return super().on_pdu(pdu)

    def inject(self, packet):
        if hasattr(packet, "address") and packet.address != self.address:
            self.address = packet.address
        elif hasattr(packet, "metadata") and hasattr(packet.metadata, "address"):
            self.address = packet.metadata.address

        if self.__configuration.synchronize:
            if not self._synced:
                if self.synchronize():
                    self.lock()
                self._injecting = True
                while not self.send(packet, channel=self.channel, address=self.address):
                    self._injecting = False
                    self.unlock()
                    if self.synchronize():
                        self.lock()
                self._synced = True
                return True
            else:
                success = self.send(packet, channel=self.channel, address=self.address)
                return success
        else:
            success = self.send(packet, channel=self.__configuration.channel, address=self.address)
